from quicktions import Fraction
from abc import ABC, abstractmethod
from typing import Union


class _VideoSource(ABC):
    """
    Abstract class that is a media source
    containing video information.
    """

    @property
    @abstractmethod
    def copy(
        self
    ):
        """
        Get a copy of this video source instance.
        """
        pass

    @property
    @abstractmethod
    def is_unchanged(
        self
    ) -> bool:
        """
        *For internal use only*

        Internal boolean flag to indicate if the source has
        been modified (or not), useful when we want to 
        optimize the way we access to the frames to render.

        If the original source has not been modified, we
        don't need to read frame by frame and apply any
        change so we can do it faster.

        (!) This property has to be modified by the specific
        video sources.
        """
        pass

    @property
    @abstractmethod
    def duration(
        self
    ):
        """
        The duration of the source.
        """
        pass

    @property
    @abstractmethod
    def size(
        self
    ):
        """
        The size of the original source, that can
        match or match not the 'output_size'.
        """
        pass

    def __init__(
        self
    ):
        pass

    @abstractmethod
    def get_video_frame_at(
        self,
        t: Union[int, float, Fraction],
        size: Union[tuple[int, int], None] = None,
        do_apply_filters: bool = True
        # TODO: Return 'VideoFrame' or 'numpy' (?)
    ):
        """
        Get the video frame that must be displayed
        at the 't' time moment.
        """
        pass

    @abstractmethod
    def get_audio_frames_at(
        self,
        t: Union[int, float, Fraction],
        video_fps: Union[int, float, Fraction],
        do_apply_filters: bool = True
        # TODO: Return 'VideoFrame' or 'numpy' (?)
    ):
        """
        Get the audio frames that must be played
        during the 't' time moment of a video.
        """
        pass

    # TODO: audio frames t (?)
        
class _AudioSource(ABC):
    """
    Abstract class that is a media source
    containing audio information.
    """

    @property
    @abstractmethod
    def copy(
        self
    ):
        """
        Get a copy of this video source instance.
        """
        pass

    @property
    @abstractmethod
    def is_unchanged(
        self
    ) -> bool:
        """
        *For internal use only*

        Internal boolean flag to indicate if the source has
        been modified (or not), useful when we want to 
        optimize the way we access to the frames to render.

        If the original source has not been modified, we
        don't need to read frame by frame and apply any
        change so we can do it faster.

        (!) This property has to be modified by the specific
        audio sources.
        """
        pass

    @property
    @abstractmethod
    def duration(
        self
    ):
        """
        The duration of the source.
        """
        pass

    def __init__(
        self
    ):
        pass

    @abstractmethod
    def get_audio_frames_at(
        self,
        t: Union[int, float, Fraction],
        video_fps: Union[int, float, Fraction],
        do_apply_filters: bool = True
        # TODO: Return 'VideoFrame' or 'numpy' (?)
    ):
        """
        Get the audio frames that must be played
        during the 't' time moment of a video.
        """
        pass




"""
TODO: I think this has to be moved from here
"""
from typing import Union


class TransitionNodeRegistry:
    """
    The registry containing the nodes we accept and
    make available in the platform.
    """

    _registry: dict = {}
    """
    *For internal use only*

    The registry itself, containing the names as keys
    and the specific `_TransitionProcessor` instances
    as values.
    """

    @classmethod
    def register(
        cls,
        name: str,
        node_processor_class: '_TransitionProcessor',
        opengl_context: Union['Context', None] = None
    ) -> None:
        """
        Register the `node_processor_class` with the given
        `name`.
        """
        cls._registry[name] = node_processor_class(
            opengl_context = opengl_context
        )

    @classmethod
    def get(
        cls,
        name: str
    ) -> '_TransitionProcessor':
        """
        Get the `node_processor_class` with the given `name`.
        """
        return cls._registry[name]

    @classmethod
    def all(
        cls
    ) -> dict[str, '_TransitionProcessor']:
        """
        Get the registry as a dict.
        """
        return dict(cls._registry)
    
    @classmethod
    def is_registered(
        cls,
        name: str
    ) -> bool:
        """
        Check if the registry with the given `name` is
        registered or not.
        """
        return cls._registry.get(name, None) is not None
    
    @classmethod
    def get_default(
        cls
    ) -> Union['_TransitionProcessor', None]:
        """
        Get the default registry, that is the first one
        if the registry contains some, or `None` if there
        are no items registered.
        """
        first_key = TransitionNodeRegistry.get_default_key()

        return (
            cls._registry[first_key]
            if first_key is not None else
            None
        )
    
    @classmethod
    def get_default_key(
        cls
    ) -> Union[str, None]:
        """
        Get the default key of the registry, that is the
        first one if the registry contains some, or `None`
        if there are no items registered.
        """
        return next(iter(cls._registry), None)

# TODO: This must be moved to an initializer, not here
def register_transition_processor_nodes():
    """
    Method to be called by the app to register the
    specific transition node processors we want to
    be implemented and available.

    These are the transitions that we have currently
    available:
    - `circle_opening`
    - `circle_closing`
    - `crossfade`
    - `distorted_crossfade`
    - `slide`
    - `bars_falling`
    - `alphapedia_mask`
    """
    from yta_editor.tracks.items.transition.transition_node_registry import TransitionNodeRegistry
    from yta_editor_nodes.processor.video.transitions import CircleOpeningTransitionProcessor, CircleClosingTransitionProcessor, CrossfadeTransitionProcessor, DistortedCrossfadeTransitionProcessor, SlideTransitionProcessor, BarsFallingTransitionProcessor, AlphaPediaMaskTransitionProcessor

    # Modify it to make more nodes available
    # TODO: This could be a json or something like this
    # and maybe according to the version of the software
    nodes = {
        'circle_opening': CircleOpeningTransitionProcessor,
        'circle_closing': CircleClosingTransitionProcessor,
        'crossfade': CrossfadeTransitionProcessor,
        'distorted_crossfade': DistortedCrossfadeTransitionProcessor,
        'slide': SlideTransitionProcessor,
        'bars_falling': BarsFallingTransitionProcessor,
        'alphapedia_mask': AlphaPediaMaskTransitionProcessor
    }

    for name, node_class in nodes.items():
        TransitionNodeRegistry.register(
            name,
            node_class,
            # TODO: What about the 'opengl_context' (?)
        )


from typing import Union


def _update_transparency(
    transparency: Union[float, None],
    format: Union[str, None] = None
) -> Union[float, None]:
    """
    *For internal use only*

    Update the provided 'transparency' value,
    that must be a value in the [0.0, 1.0] range,
    if necessary.

    If the 'format' is provided:
    - Alpha format will force an opaque alpha
    transparency if it is None
    - Non-alpha format will force a None value
    if transparency is provided
    """
    return (
        # Alpha format, no transparency => opaque
        0.0
        if (
            transparency is None and
            format is not None and
            'a' in format 
        ) else
        # Non-alpha format but transparency => None
        None
        if (
            transparency is not None and
            format is not None and
            'a' not in format
        ) else
        # Lower than 0.0 limit
        0.0
        if (
            transparency is not None and
            transparency < 0.0
        ) else
        # Greater than 1.0 limit
        1.0
        if (
            transparency is not None and
            transparency > 1.0
        ) else
        transparency
    )

def _transparency_to_alpha(
    transparency: Union[float, None]
) -> Union[int, None]:
    """
    *For internal use only*

    The `transparency` value provided must be a
    None or a value in the `[0.0, 1.0]` range,
    meaning  this:
    - `0.0`: No transparency, full color
    - `1.0`: Full transparent, no color

    The `alpha` value returned is the one that
    will be used as the alpha channel of an RGBA
    color, in which 255 means the full presence
    of the color, meaning no transparency, and 0
    the absence of color, meaning that it is full
    transparent.

    The formula:
    - `int(255 - transparency * 255)`
    """
    return (
        int(255 - transparency * 255)
        if transparency is not None else
        transparency
    )


from yta_editor.tracks.items.transition.transition_node_registry import TransitionNodeRegistry
from yta_editor.tracks.items.transition.transition_specification import TransitionSpecification
from yta_editor.media.video import VideoFileMedia
from yta_editor_common.frame.helper import FrameHelper
from yta_editor_utils.texture import TextureUtils
from yta_validation import PythonValidator
from yta_math_easings.abstract import EasingFunction
from av.video.frame import VideoFrame
from quicktions import Fraction
from typing import Union

import numpy as np


# TODO: This class below could be repeated and I just
# created because I needed to (at least by now)
# TODO: The progress is not being calculated correctly
class _TransitionCalculator:
    """
    *For internal use only*

    A class to simplify the way we make the calculations
    related to a transition. This class must be used within
    a transition item that needs these kind of calculations.
    """

    @property
    def fps(
        self
    ) -> float:
        """
        The `fps` associated to the `transition_item` that
        is attached to this instance.
        """
        # TODO: Is this the 'output_fps' (?)
        return self._transition_item.fps
    
    @property
    def output_size(
        self
    ) -> float:
        """
        The `output_size` associated to the `transition_item`
        that is attached to this instance.
        """
        return self._transition_item.output_size

    def __init__(
        self,
        transition_item: 'TransitionTrackItem',
        duration: float,
        transition_specification: Union['TransitionSpecification', None] = None,
        do_use_gpu: bool = True,
    ):
        self._transition_item: 'TransitionTrackItem' = transition_item
        """
        The transition item this helper was instantiated for.
        """
        self.duration: float = duration
        """
        The duration of the transition.
        """
        self._do_use_gpu: bool = do_use_gpu
        """
        *For internal useo only*

        The boolean flag to indicate if the GPU must be used
        or not when applying the node associated to this
        transition.
        """

        transition_specification = (
            TransitionSpecification(
                node_name = TransitionNodeRegistry.get_default_key(),
                # TODO: This must be dynamic according to the default
                parameters = {
                    'border_smoothness': 0.02
                },
                easing_function = 'linear'
            )
            if transition_specification is None else
            transition_specification
        )

        if not TransitionNodeRegistry.is_registered(transition_specification.node_name):
            raise Exception(f'There is no transition node processor registered with the given "{transition_specification.node_name}"')
        
        # TODO: The parameters of the specification must be
        # validated here according to the node processor
        # specification (which doesn't exist yet). By now
        # we are not doing this parameters validation
        self._transition_specification: 'TransitionSpecification' = transition_specification
        """
        The specification that defines the node processor
        and the parameters we should use.
        """

        # The transition could need the 'mask_input'
        mask_input_filename = self._transition_specification.parameters.get('mask_input', None)
        self._mask_video: Union[VideoFileMedia, None] = (
            VideoFileMedia(
                filename = mask_input_filename
            )
            if mask_input_filename is not None else
            None
        )
        """
        *For internal use only*

        The alpha video that we will use to read the
        frames and apply the mask.
        """
        self._node_processor = TransitionNodeRegistry.get(self._transition_specification.node_name)
        # I think it should be initialized here...
        # (
        #     # TODO: What do I do with the 'opengl_context' (?)
        #     opengl_context = None
        # )
        """
        *For internal use only*

        The node processor that we will use to process
        the input frame with the corresponding mask
        frames.
        """

    # TODO: This method can be overwritten to change
    # its behaviour if the specific transition needs
    # it
    def process_frame(
        self,
        frame_a: Union['moderngl.Texture', 'np.ndarray'],
        frame_b: Union['moderngl.Texture', 'np.ndarray'],
        evaluation_context: 'EvaluationContext'
    ) -> VideoFrame:
        """
        Process the `frame_a` and the `frame_b` provided and
        apply the transition at the `t_progress` provided by
        using the node of the `TransitionTrackItem` associated
        to this instance.
        """
        # TODO: Maybe this can be placed in the general
        # class if it doesn't change

        # TODO: This can be a texture
        processed_frame = self._get_process_frame_array(
            frame_a = frame_a,
            frame_b = frame_b,
            evaluation_context = evaluation_context
        # TODO: Force this with the 'yta-numpy' utils to 'np.uint8'
        )

        processed_frame = frame_to_uint8_ndarray(
            frame = processed_frame,
            do_include_alpha = False
        )

        # if PythonValidator.is_numpy_array(processed_frame):
        #     processed_frame = processed_frame.astype(np.uint8)

        #     # frame = VideoFrame.from_ndarray(
        #     #     # TODO: Force this with the 'yta-numpy' utils to 'np.uint8'
        #     #     array = processed_frame,
        #     #     format = 'rgb24'
        #     # )
        # elif PythonValidator.is_instance_of(processed_frame, 'Texture'):
        #     processed_frame, has_alpha = FrameHelper.texture_to_ndarray(
        #         frame = processed_frame
        #     )
            
        #     # TODO: Removing the alpha channel here but, should I (?)
        #     if has_alpha:
        #         processed_frame = processed_frame[:, :, :3]

        # else:
        #     # TODO: Wtf? This should not happen
        #     raise Exception('Wowowowooo, no texture, no np.ndarray... wtf?')
        
        frame = FrameHelper.ndarray_to_videoframe(
            frame = processed_frame,
            do_include_alpha = False,
            pts = None,
            time_base = Fraction(1, self.fps)
        )
        
        # TODO: If I return this as a numpy and make the
        # transformations and then we want to use that
        # same frame within other GPU processor... we are
        # f***
        return frame
    
    # TODO: Overwrite this method to make it custom (?)
    def _get_process_frame_array(
        self,
        frame_a: Union['moderngl.Texture', 'np.ndarray'],
        frame_b: Union['moderngl.Texture', 'np.ndarray'],
        evaluation_context: 'EvaluationContext'
    ) -> Union[np.ndarray, 'moderngl.Texture']:
        """
        Get the numpy array that will be used to build the
        pyav VideoFrame that will be returned.
        """
        inputs = {
            'first_input': frame_a,
            'second_input': frame_b
        }

        # TODO: Check this below to improve it if possible
        parameters_dynamic = self._transition_specification.parameters.copy()

        # Alpha mask transition, we need to handle it in
        # a special way
        if self._mask_video is not None:
            # TODO: Make it a property (?)
            easing_function = EasingFunction.get(self._transition_specification.easing_function)()
            # We calculate the transition t_media with the progress
            t_transition_media = easing_function.ease(evaluation_context.progress.value) * (self._mask_video.number_of_frames - 1) / self._mask_video.fps

            # We get a VideoFrame, so we transform it
            # TODO: Maybe transform it in other way for optimization
            inputs['mask_input'] = self._mask_video.get_video_frame_at(
                t_media = t_transition_media
            ).to_ndarray(format = 'rgba')

            # We don't want it as a dynamic parameter but as input
            del parameters_dynamic['mask_input']

        return self._node_processor.process(
            inputs = inputs,
            evaluation_context = evaluation_context,
            # TODO: This is how we were using the easing function before
            # progress = self.easing_function.ease(float(evaluation_context.progress)),
            output_size = self.output_size,
            do_use_gpu = self._do_use_gpu,
            # Parameters will be validated in
            parameters = parameters_dynamic
        )
    
    def _get_process_audio_frames(
        self,
        audio_frames_a: list['AudioFrameWrapped'],
        audio_frames_b: list['AudioFrameWrapped'],
        evaluation_context: 'EvaluationContext'
    ) -> list['AudioFrame']:
        """
        Get the numpy array that will be used to build the
        pyav AudioFrame that will be returned.
        """

        """
        TODO: I don't know how to handle the audio within the
        transitions, because each transition works in a specific
        way and making the audio behave like the video movement
        is not easy... It could be based on the 't_progress',
        using something like this below (based on S-Curve Crossfade
        weights):

        weight_a = 0.5 * (1 + math.cos(t_progress * math.pi))
        weight_b = 1 - weight_a
        mixed_audio_frames = audio_frames_a * weight_a + audio_frames_b * weight_b

        But, as I don't know actually how to do it, by now we are
        just mixing both of them during the transition time in the
        same way as the video frames are accessed.
        """
        # TODO: Move this to top (?)
        from yta_editor.utils.frame_combinator import AudioFrameCombinator
        from yta_editor.timeline import concatenate_audio_frames

        collapsed_frames = [
            concatenate_audio_frames(frames)
            for frames in [audio_frames_a, audio_frames_b]
        ]

        # We keep only the non-silent frames because
        # we will sum them after and keeping them will
        # change the results.
        non_empty_collapsed_frames = [
            frame._frame
            for frame in collapsed_frames
            if not frame.is_from_gap
        ]

        if len(non_empty_collapsed_frames) == 0:
            # If they were all silent, just keep one
            non_empty_collapsed_frames = [collapsed_frames[0]._frame]

        frames = [
            AudioFrameCombinator.sum_tracks_frames(
                tracks_frames = non_empty_collapsed_frames,
                sample_rate = self._transition_item._track.audio_fps,
                # TODO: This was not being sent before
                layout = self._transition_item._track.audio_layout,
                format = self._transition_item._track.audio_format
            )
        ]

        # TODO: Do we need the '**self._transition_specification.parameters'
        # for the audio frames (?)

        # for audio_frame in frames:
        #     yield audio_frame
        # TODO: Should I yield (?)
        return frames


def frame_to_uint8_ndarray(
    frame: Union['moderngl.Texture', 'np.ndarray'],
    do_include_alpha: bool = True
):
    """
    Force the `frame` provided to be transformed into
    a uint8 numpy ndarray, removing the alpha channel
    if existing and `do_include_alpha` is `False`.
    """
    if (
        not PythonValidator.is_instance_of(frame, 'Texture') and
        not PythonValidator.is_numpy_array(frame)
    ):
        # TODO: Wtf? This should not happen
        raise Exception('Wowowowooo, no texture, no np.ndarray... wtf?')
    
    if PythonValidator.is_instance_of(frame, 'Texture'):
        frame, has_alpha = FrameHelper.texture_to_ndarray(frame)

        if (
            has_alpha and
            not do_include_alpha
        ):
            frame = frame[:, :, :3]

    return TextureUtils.numpy_to_uint8(frame)


from yta_editor.transformations.effects.abstract import Effect
from yta_editor.utils.transition_calculator import frame_to_uint8_ndarray
from yta_editor_time.evaluation_context import EvaluationContext
# from av.video import VideoFrame
from typing import Union
from abc import abstractmethod

import numpy as np


class VideoEffect(Effect):
    """
    An effect for a video input.

    This class is just to be able to identify the
    hierarchy and which effect is designed for 
    the video.
    """

    @abstractmethod
    def _apply(
        self,
        # TODO: Set the type
        frame: any,
        output_size: Union[tuple[int, int], None],
        evaluation_context: EvaluationContext
    # TODO: Set the type
    ) -> any:
        """
        *For internal use only*

        The internal method to apply the effect on the `frame`
        provided.
        """
        pass

    def apply(
        self,
        # TODO: Set the type
        frame: any,
        output_size: Union[tuple[int, int], None],
        evaluation_context: EvaluationContext
    # TODO: Set the type
    ) -> any:
        """
        Apply the effect to the given `frame` at the `t` time
        moment provided.
        """
        frame_processed = self._apply(
            frame = frame,
            output_size = output_size,
            evaluation_context = evaluation_context
        )

        frame_processed = frame_to_uint8_ndarray(
            frame = frame_processed,
            do_include_alpha = True
        )

        # TODO: Why do we force it to be a numpy ndarray (?)
        # The effect can be done using GPU

        # if PythonValidator.is_instance_of(frame_processed, 'Texture'):
        #     from yta_editor_common.frame.helper import FrameHelper
        #     frame_processed = FrameHelper.texture_to_ndarray(frame_processed)[0]

        # frame_processed = TextureUtils.numpy_to_uint8(frame_processed)

        # The output is forced to be 'uint8'
        return frame_processed

class VideoEffects:
    """
    A set of effects to apply in a video input.
    """

    # TODO: This is not as simple as a list, it must be
    # managed with the specific nodes we designed that
    # are able to combine and are time-concerned.

    def __init__(
        self,
        effects: list[VideoEffect] = []
    ):
        self.effects: list[VideoEffect] = effects
        """
        The list of the effects to apply in the video.
        """

    def apply(
        self,
        frame: Union['np.ndarray', 'Texture'],
        output_size: Union[tuple[int, int], None],
        evaluation_context: EvaluationContext
    # TODO: Set the type
    ) -> Union['np.ndarray', 'Texture']:
        """
        Apply all the effects that should be applied at the
        given `t` time moment to the provided audio `frame`.
        """
        # width, height = frame.width, frame.height
        # Frame here: 'ndarray' (CPU) or 'Texture' (GPU)

        if len(self.effects) > 0:
            # TODO: Only if 'frame' is 'av.VideoFrame'
            # processed_frame = frame.to_ndarray(format = 'rgba')
            processed_frame = frame

            for effect in self.effects:
                # TODO: The effect has to be a real input (numpy or texture)
                processed_frame = effect.apply(
                    frame = processed_frame,
                    output_size = output_size,
                    evaluation_context = evaluation_context
                )

            # TODO: Only if texture:
            # data = processed_frame.read()
            # data = np.frombuffer(data, dtype = np.uint8)
            # # The shape has to be the expected by the timeline
            # # TODO: Do I need the 'output_size' here? It should be
            # # defined before the 'transform' processing...
            # # data = data.reshape(self.output_size[1], self.output_size[0], 4)
            # data = data.reshape(height, width, 4)
            # # OpenGL uses textures flipped, so...
            # data = np.flip(data, axis = 0)
            # # TODO: I'm removing the alpha channel here...
            # # processed_frame = data[:, :, :3]

            # If numpy
            # data = processed_frame.astype(np.uint8)
            
            # frame = VideoFrame.from_ndarray(
            #     # TODO: Force this with the 'yta-numpy' utils to 'np.uint8'
            #     array = data,
            #     #format = 'rgb24'
            #     format = 'rgba'
            # )

            return processed_frame

        return frame



from yta_editor.tracks.items.abstract import _TrackItem
from yta_time_interval import TIME_INTERVAL_SYSTEM_LIMITS
from quicktions import Fraction
from typing import Union


# TODO: This is repeated and not used everywhere
Number = Union[int, float, Fraction]
"""
Numeric type we accept as parameter.
"""
# TODO: What about the inheritance (?)
# TODO: Replace the 'empty' video and audio logic with
# this class
class GapTrackItem(_TrackItem):
    """
    A simple class to replace the logic related to 'empty'
    video track items and be more simple and understandable.

    This item will be placed when there is a gap in between
    2 real clips.
    """

    @property
    def _is_last_gap(
        self
    ) -> bool:
        """
        *For internal use only*

        Flag to indicate that the gap instance is the last one
        in the track, which means that the 't_end' is the system's
        limit of
        `yta_video_frame_time.interva.TIME_INTERVAL_SYSTEM_LIMITS`.
        """
        return self.t_end == TIME_INTERVAL_SYSTEM_LIMITS[1]

    @property
    def copy(
        self
    ) -> 'GapTrackItem':
        """
        A copy of the current gap instance.
        """
        return GapTrackItem(
            track = self._track,
            t_start = self.t_start,
            duration = self.duration,
            item_in = self.item_in,
            item_out = self.item_out
        )
    
    @property
    def is_unchanged(
        self
    ) -> bool:
        """
        *For internal use only*

        Internal boolean flag to indicate if the source has
        been modified (or not), useful when we want to 
        optimize the way we access to the frames to render.

        If the original source has not been modified, we
        don't need to read frame by frame and apply any
        change so we can do it faster.

        (!) This property is useless here as it is a gap.
        """
        return True

    def __init__(
        self,
        track: 'Track',
        t_start: float,
        duration: float,
        item_in: Union['_AudioTrackItem', '_VideoTrackItem', 'TransitionTrackItem', 'GapTrackItem', None] = None,
        item_out: Union['_AudioTrackItem', '_VideoTrackItem', 'TransitionTrackItem', 'GapTrackItem', None] = None
    ):
        # TODO: Maybe use the **kwargs (?)
        super().__init__(
            track = track,
            t_start = t_start,
            duration = duration,
            item_in = item_in,
            item_out = item_out,
            transition_in = None,
            transition_out = None
        )

    """
    The gap track item is special because it can be the
    last item of the track and have an t_end that is the
    system's t_end value and that 't_end' value must be kept
    as it is, so when manipulating the time interval we
    have to do it carefully and in an special way using
    the internal `_is_last_gap` flag.
    """

    def _shift_by(
        self,
        delta: Number
    ) -> '_TrackItem':
        """
        Move the `t_start` and `t_end` value of this item the amount
        of time defined by the `delta` provided parameter, that
        can be positive or negative (and will be truncated to be
        a multiple of `1/fps`).
        """
        if self._is_last_gap:
            return self.shift_start_by(delta)
        
        self._time_interval.shift_by(delta)
        
        return self
    
    def shift_to(
        self,
        t_track: Number
    ) -> '_TrackItem':
        """
        Move the `t_start` and `t_end` value of this item to the time
        moment defined by the `t_track` parameter provided. The
        change will not modify the duration of the item.
        """
        if self._is_last_gap:
            return self.shift_start_to(t_track)
        
        self._time_interval.shift_by(
            delta = t_track - self.t_start
        )
        
        return self
    
    def shift_start_by(
        self,
        delta: Number
    ) -> '_TrackItem':
        """
        Update the `t_start` value by adding the `delta` provided
        (that can be positive to make it start later or negative
        to make it start earlier).
        """
        self._time_interval.shift_start_by(delta)

        return self
    
    def shift_start_to(
        self,
        t_track: Number
    ) -> '_TrackItem':
        """
        Update the `t_start` value to the one provided as `t_track`
        (that can be before or after the current one).
        """
        return self.shift_start_by(
            delta = t_track - self.t_start
        )
    
    def shift_end_by(
        self,
        delta: Number
    ) -> '_TrackItem':
        """
        Update the `t_end` value by adding the `delta` provided
        (that can be positive to make it start later or negative
        to make it start earlier).
        """
        if self._is_last_gap:
            return self
        
        self._time_interval.shift_end_by(delta)

        return self
    
    def shift_end_to(
        self,
        t_track: Number
    ) -> '_TrackItem':
        """
        Update the `t_end` value to the one provided as `t_track`
        (that can be before or after the current one).
        """
        if self._is_last_gap:
            return self
        
        return self.shift_end_by(
            delta = t_track - self.t_end
        )
    
    def _get_audio_frames_at(
        self,
        t_track: Union[int, float, Fraction],
        video_fps: Union[int, float, Fraction, None] = None,
        do_use_source_limits: bool = False,
        do_apply_filters: bool = True,
    ):
        """
        Get the sequence of audio frames for the 
        given video 't' time moment, using the
        audio cache system.

        The `t` time moment must be a value between
        0 and the media duration, that will be
        transformed into the corresponding source
        time moment to be read.

        This is useful when we want to write a
        video frame with its audio, so we obtain
        all the audio frames associated to it
        (remember that a video frame is associated
        with more than 1 audio frame).
        """
        return self._track.audio_silent_from_gap
    
    def _get_video_frame_at(
        self,
        t_track: Union[int, float, Fraction],
        do_use_source_limits: bool = False
    ) -> Union['VideoFrameWrapped', None]:
        """
        Get the video frame for the provided `t_track` time
        moment, that will be always a black frame because 
        this is a gap instance.

        The empty frames we need to send to the editor
        as the content when we don't have any frame to
        show, that can be because we have a gap (black
        screen) in that specific moment.

        This property has been created to avoid
        creating these empty frames again and again,
        as they will be always the same.
        """
        return self._track.video_empty_frame


from yta_constants.enum import YTAEnum as Enum


class TransitionMode(Enum):
    """
    The modes the TransitionTrackItem accept as mode to
    implement it, that is the way you handle the frames
    of the previous and the next clip.
    """

    TRIM = 'trim'
    """
    The transition will use the real frames of both the
    previous and the next clip, which will make both 
    clips last less time.

    It means that from:
    - `[ClipA] [Gap] [ClipB]`
    We will move to:
    - `[ClipA_trimmed] [Transition] [ClipB_trimmed]`
    """
    FREEZE_TAIL = 'freeze_tail'
    """
    The transition will use the last frame of the 
    previous clip but frozen, which means that it will
    be used for the whole transition duration. The next
    clip 'start' time moment will be increased because
    the previous clip is not trimmed and the transition
    will last an additional amount of time.

    It means that from:
    - `[ClipA] [Gap] [ClipB]`
    We will move to:
    - `[ClipA] [Transition] [ClipB_delayed]`
    """
    FREEZE_HEAD = 'freeze_head'
    """
    The transition will use the first frame of the next
    clip but frozen, which means that it will be used
    for the whole transition duration. The next clip
    'start' time moment will be increased because the
    previous clip is not trimmed and the transition will
    last an additional amount of time.

    It means that from:
    - `[ClipA] [Gap] [ClipB]`
    We will move to:
    - `[ClipA] [Transition] [ClipB_delayed]`
    """
    FREEZE_BOTH = 'freeze_both'
    """
    The transition will use the last frame of the
    previous clip and the first frame of the next one,
    but frozen, which means that they will be used for
    the whole transition duration. The next clip 'start'
    time moment will be increased because the previous
    clip is not trimmed and the transition will last an
    additional amount of time.

    It means that from:
    - `[ClipA] [Gap] [ClipB]`
    We will move to:
    - `[ClipA] [Transition] [ClipB_delayed]`
    """


# from yta_numpy.rgba.utils import RGBAFrameUtils
# from yta_editor_nodes.old_opengl.effects import AudioEffectsStack, VideoEffectsStack
# from yta_validation import PythonValidator
# from yta_validation.parameter import ParameterValidator
# from av.video.frame import VideoFrame
# from av.audio.frame import AudioFrame
# from quicktions import Fraction
# from typing import Union


# # TODO: These methods below are old version and
# # now we are implementing effects through nodes
# """
# These methods below are shared by the
# Audio and Video class that handle and
# wrap an audio or video.
# """
# def apply_video_effects_to_frame_at(
#     effects_stack: VideoEffectsStack,
#     frame: VideoFrame,
#     t: Union[int, float, 'Fraction']
# ) -> Union[VideoFrame, 'ndarray']:
#     """
#     Apply the video effects to the given
#     'frame' on the 't' time moment provided.

#     This method should be called before 
#     yielding any frame.
#     """
#     ParameterValidator.validate_mandatory_instance_of('frame', frame, VideoFrame)
#     ParameterValidator.validate_mandatory_instance_of('effects_stack', effects_stack, VideoEffectsStack)

#     if len(effects_stack.effects) == 0:
#         return frame

#     # TODO: I think this has to preserve the
#     # transparency if any of the video frames
#     # that are passed to this method (from 
#     # the different layers) have an alpha
#     # channel or we will lose it in the
#     # process... By now I'm just forcing the
#     # format that comes. Or maybe not,
#     # because when a frame comes and has no
#     # transparency, the next frames will be
#     # not visible (if coming ordered by
#     # priority)... Well... maybe I'm confused
#     # right now and this just modifies a 
#     # single frame or a frame that has been
#     # combined previously, so no worries...

#     # We handle frames as 'rgb24' or 'rgba' and
#     # then reformat to the expected format (if
#     # needed)
#     temp_format = (
#         'rgba'
#         if (
#             frame.format.name != 'rgba' and
#             'a' in frame.format.name
#         ) else
#         'rgb24'
#         if (
#             frame.format.name != 'rgb24' and
#             'a' not in frame.format.name
#         ) else
#         frame.format.name
#     )

#     # Need to send the frame as a numpy for
#     # the effects
#     new_frame = effects_stack.apply_effects_at(
#         frame = frame.to_ndarray(
#             format = temp_format
#         ),
#         # The 't' here is the internal valid one
#         t = t
#     )

#     """
#     When applying the video effects we use
#     opengl textures that return, always, an
#     alpha channel, so we need to remove it
#     if we don't actually need it
#     """
#     new_frame = (
#         RGBAFrameUtils.remove_alpha_channel_from_frame(new_frame)
#         if temp_format == 'rgb24' else
#         new_frame
#     )

#     # Rebuild the VideoFrame
#     new_frame = VideoFrame.from_ndarray(
#         array = new_frame,
#         format = temp_format
#     )

#     new_frame = (
#         new_frame.reformat(format = frame.format.name)
#         if frame.format.name != temp_format else
#         new_frame
#     )

#     """
#     We need 'time_base' and 'pts' values to
#     be identified by pyav as valid frames 
#     but we don't actually care about the
#     values because the Timeline that renders
#     will overwrite them
#     """
#     new_frame.time_base = (
#         frame.time_base
#         if frame.time_base is not None else
#         Fraction(1, 60)
#     )

#     new_frame.pts = (
#         frame.pts
#         if frame.pts is not None else
#         0
#     )

#     return new_frame

# def apply_audio_effects_to_frame_at(
#     effects_stack: AudioEffectsStack,
#     frame: Union['AudioFrame', 'ndarray'],
#     t: Union[int, float, 'Fraction']
# ) -> Union['AudioFrame', 'ndarray']:
#     """
#     Apply the audio effects to the given
#     'frame' on the 't' time moment provided.

#     This method should be called before 
#     yielding any frame.
#     """
#     # TODO: I think we shouldn't receive a
#     # 'ndarray' here, it must be AudioFrame
#     ParameterValidator.validate_mandatory_instance_of('frame', frame, [AudioFrame, 'ndarray'])

#     # Need the frame as a numpy
#     new_frame = (
#         frame.to_ndarray()
#         if PythonValidator.is_instance_of(frame, AudioFrame) else
#         frame
#     )
    
#     new_frame = effects_stack.apply_effects_at(
#         frame = new_frame,
#         # The 't' here is the internal valid one
#         t = t
#     )

#     # Rebuild the AudioFrame
#     new_frame = AudioFrame.from_ndarray(
#         array = new_frame,
#         format = frame.format,
#         layout = frame.layout
#     )

#     new_frame.sample_rate = frame.sample_rate

#     """
#     When applying the video effects we use
#     opengl textures that return, always, an
#     alpha channel, so we need to remove it
#     if we don't actually need it
#     """
#     new_frame.time_base = (
#         frame.time_base
#         if frame.time_base is not None else
#         Fraction(1, 60)
#     )

#     new_frame.pts = (
#         frame.pts
#         if frame.pts is not None else
#         0
#     )

#     return new_frame




from yta_editor.tracks.items.abstract import TrackItemWithAudioMedia, _TrackItemWithVideo
from yta_validation.parameter import ParameterValidator
from quicktions import Fraction
from typing import Union


# TODO: Having the 'TrackItemWithVideoMedia' I don't need this
class _VideoTrackItem(TrackItemWithAudioMedia, _TrackItemWithVideo):
    """
    Class to represent an element that is on the
    track, that can be an empty space or a video.
    """

    def __init__(
        self,
        track: 'VideoTrack',
        t_start: Union[int, float, Fraction],
        t_end: Union[int, float, Fraction],
        media: Union['VideoTimed', None] = None,
        video_transform: Union['VideoTransform', None] = None,
        video_effects: Union['VideoEffects', None] = None,
        item_in: Union['_VideoTrackItem', 'TransitionTrackItem', 'GapTrackItem', None] = None,
        item_out: Union['_VideoTrackItem', 'TransitionTrackItem', 'GapTrackItem', None] = None
    ):
        ParameterValidator.validate_instance_of('media', media, 'VideoTimed')

        super().__init__(
            track = track,
            t_start = t_start,
            t_end = t_end,
            media = media,
            video_transform = video_transform,
            video_effects = video_effects,
            item_in = item_in,
            item_out = item_out
        )

        self.transition_in: Union['TransitionTrackItem'] = None
        """
        The transition that will make the previous video in the
        track be mixed with this one at the begining.
        """
        self.transition_out: Union['TransitionTrackItem'] = None
        """
        The transition that will make the next video in the
        track be mixed with this one at the t_end.
        """


from yta_editor_nodes import TimelineGraph, SerialTimelineNode
from yta_editor_nodes.compositor import DisplacementWithRotationNodeCompositor
from yta_editor_parameters.values.abstract import VideoEditorParameterValue
from yta_editor_parameters.values import ConstantVideoEditorParameterValue
from yta_editor_parameters.values.utils import parse_parameter, is_this_constant_value
from yta_editor_time.evaluation_context import EvaluationContext
from yta_editor_time.specifications import FrameIndexSpecification, ProgressSpecification, TimeSpecification
from yta_validation.parameter import ParameterValidator
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Union

import numpy as np


@dataclass
class ParameterValueUnset:
    """
    A dataclass just to indicate that a parameter value
    has not been set but is not None.
    """

    pass

VALUE_UNSET = ParameterValueUnset()
"""
A parameter value that has not been set yet and
is different than being `None`.
"""


@dataclass
class AudioTransformState:
    """
    *Dataclass*

    A dataclass to represent the information about the
    state of a transformation we should apply to 
    transform an audio, that has been calculated for a
    specific `t` time moment.
    """

    def __init__(
        self,
        volume: float = 0.0
    ):
        self.volume: float = volume
        """
        The volume to apply at the audio at that moment.
        """

class Transform(ABC):
    """
    *Abstract class*

    Abstract class to be inherited by the `Transform`
    implementation related to audio and video.
    """

    @abstractmethod
    def apply(
        self,
        # TODO: Set the type
        frame: any,
        evaluation_context: EvaluationContext
    # TODO: Set the type
    ) -> any:
        """
        Apply the transformation to the given `frame` at the also
        provided `t` (which will be used to obtain the transform
        state to apply).
        """
        pass

class VideoTransform(Transform):
    """
    The transformation conditions to define how we want
    to modify the different time moments of a video in
    the timeline.

    This is attached to a timeline video track item and
    will be applied on it. This is useful to apply the
    basic modifications.

    These values are time functions, and the values
    change according to the `t` time moment in which
    the values are requested: position, rotation,
    scale, opacity, etc.

    Giving not `width` and/nor `height` will make the
    system use the size of the media associated (if
    available and possible) to keep its original size.

    TODO: I think we should have 't_start' and 't_end'
    for another version that is applied on the track
    instead.
    """

    @property
    def copy(
        self
    ) -> 'VideoTransform':
        """
        Get a copy of it.
        """
        return VideoTransform(
            # TODO: Maybe we need a '.copy' of these below (?)
            specification = self.specification,
            output_size = self.output_size,
            x = self.x,
            y = self.y,
            width = self.width,
            height = self.height,
            rotation = self.rotation,
            opacity = self.opacity,
        )
    
    @property
    def timeline_graph(
        self
    ) -> TimelineGraph:
        """
        The `TimelineGraph` to make the changes, that is
        created when needed.
        """
        if self._timeline_graph is None:
            self._timeline_graph = self._build_timeline_graph()

        return self._timeline_graph
    
    @property
    def output_size(
        self
    ) -> tuple[int, int]:
        """
        The size that must be used to render the frames as the
        final output.
        """
        return self._output_size
    
    @output_size.setter
    def output_size(
        self,
        value
    ):
        self._output_size = value
        self._invalidate_graph()
    
    @property
    def x(
        self
    ) -> VideoEditorParameterValue:
        """
        The `x` position of the frame on the canvas, which is
        attached to the center of the item.
        """
        return self._x
    
    @x.setter
    def x(
        self,
        value
    ) -> None:
        self._x = value
        self._invalidate_graph()

    @property
    def y(
        self
    ) -> VideoEditorParameterValue:
        """
        The `y` position of the frame on the canvas, which is
        attached to the center of the item.
        """
        return self._y
    
    @y.setter
    def y(
        self,
        value
    ) -> None:
        self._y = value
        self._invalidate_graph()

    @property
    def width(
        self
    ) -> VideoEditorParameterValue:
        """
        The `width` of the item when displayed on the canvas.
        This value can be different than the width of the
        `output_size` and could be filled with transparent
        pixels to reach the `output_size`.
        """
        return self._width
    
    @width.setter
    def width(
        self,
        value
    ) -> None:
        self._width = value
        self._invalidate_graph()

    @property
    def height(
        self
    ) -> VideoEditorParameterValue:
        """
        The `height` of the item when displayed on the canvas.
        This value can be different than the height of the
        `output_size` and could be filled with transparent
        pixels to reach the `output_size`.
        """
        return self._height
    
    @height.setter
    def height(
        self,
        value
    ) -> None:
        self._height = value
        self._invalidate_graph()

    @property
    def rotation(
        self
    ) -> VideoEditorParameterValue:
        """
        The `rotation` of the item on the canvas.
        """
        return self._rotation
    
    @rotation.setter
    def rotation(
        self,
        value
    ) -> None:
        self._rotation = value
        self._invalidate_graph()

    @property
    def opacity(
        self
    ) -> VideoEditorParameterValue:
        """
        The output size we want to be able to reshape the frame
        after the transformation, that must fit the output of
        the track in which the item is placed.
        """
        return self._opacity
    
    @opacity.setter
    def opacity(
        self,
        value
    ) -> None:
        self._opacity = value
        self._invalidate_graph()
    
    def __init__(
        self,
        specification: Union[FrameIndexSpecification, ProgressSpecification, TimeSpecification],
        output_size: Union[tuple[int, int], ParameterValueUnset] = VALUE_UNSET,
        do_use_gpu: bool = True,
        opengl_context: Union['moderngl.Context', None] = None,
        x: Union[VideoEditorParameterValue, ParameterValueUnset] = ConstantVideoEditorParameterValue(0),
        y: Union[VideoEditorParameterValue, ParameterValueUnset] = ConstantVideoEditorParameterValue(0),
        width: Union[VideoEditorParameterValue, ParameterValueUnset] = VALUE_UNSET,
        height: Union[VideoEditorParameterValue, ParameterValueUnset] = VALUE_UNSET,
        rotation: Union[VideoEditorParameterValue, ParameterValueUnset] = ConstantVideoEditorParameterValue(0),
        opacity: Union[VideoEditorParameterValue, ParameterValueUnset] = ConstantVideoEditorParameterValue(1.0),
    ):
        # THe parameters are validated when parsed
        self.specification: Union[FrameIndexSpecification, ProgressSpecification, TimeSpecification] = specification
        """
        The specification of when the `VideoTransform` must be
        applied.
        """
        self._output_size: tuple[int, int] = output_size
        """
        The size that must be used to render the frames as the
        final output.
        """
        self.do_use_gpu: bool = do_use_gpu
        """
        Boolean flag to indicate if the nodes must use the GPU
        when rendering or if not.
        """
        self.opengl_context: Union['moderngl.Context', None] = opengl_context
        """
        The OpenGL context to be used if the transformation is
        based on the GPU.
        """
        self._x: VideoEditorParameterValue = x
        """
        The `x` position of the frame on the canvas, which is
        attached to the center of the item.
        """
        self._y: VideoEditorParameterValue = y
        """
        The `y` position of the frame on the canvas, which is
        attached to the center of the item.
        """
        self._width: VideoEditorParameterValue = width
        """
        The `width` of the item when displayed on the canvas.
        This value can be different than the width of the
        `output_size` and could be filled with transparent
        pixels to reach the `output_size`.
        """
        self._height: VideoEditorParameterValue = height
        """
        The `height` of the item when displayed on the canvas.
        This value can be different than the height of the
        `output_size` and could be filled with transparent
        pixels to reach the `output_size`.
        """
        self._rotation: VideoEditorParameterValue = rotation
        """
        The `rotation` of the item on the canvas.
        """
        self._opacity: VideoEditorParameterValue = opacity
        """
        The output size we want to be able to reshape the frame
        after the transformation, that must fit the output of
        the track in which the item is placed.
        """

        self._timeline_graph: Union[TimelineGraph, None] = None
        """
        *For internal use only*

        The internal timeline graph to be able to apply
        the transform changes.
        """

    def _invalidate_graph(
        self
    ) -> 'VideoTransform':
        """
        Set the internal `timeline_graph` variable to None
        to force it to be rebuilt with the parameters that
        are set in this instance.
        """
        self._timeline_graph = None

        return self

    def _build_timeline_graph(
        self
    ) -> 'VideoTransform':
        """
        *For internal use only*

        Build the `TimelineGraph` instance with the parameters
        provided.
        """
        # Parse parameters
        self.x = parse_parameter(self.x)
        self.y = parse_parameter(self.y)
        self.width = parse_parameter(self.width)
        self.height = parse_parameter(self.height)
        self.rotation = parse_parameter(self.rotation)
        self.opacity = parse_parameter(self.opacity)

        # TODO: We should validate 'output_size' as tuple
        if self.output_size is VALUE_UNSET:
            raise Exception('The "output_size" provided is UNSET')

        self._timeline_graph = TimelineGraph(
            root_node = SerialTimelineNode(
                node = DisplacementWithRotationNodeCompositor(
                    # TODO: Where do I have the context (?)
                    opengl_context = None
                ),
                name = 'video_transform_base',
                specification = self.specification,
                parameters = {
                    'x': self.x,
                    'y': self.y,
                    'width': self.width,
                    'height': self.height,
                    'rotation': self.rotation
                }
            )
        )

        return self

    def apply(
        self,
        frame: Union['np.ndarray', 'Texture'],
        evaluation_context: EvaluationContext
    ) -> Union['np.ndarray', 'Texture']:
        """
        Apply the transformation to the given `frame` at the also
        provided `t` (which will be used to obtain the transform
        state to apply).
        """
        if self._timeline_graph is None:
            self._build_timeline_graph()

        # We process only if there are changes to apply
        if (
            not is_this_constant_value(self.x, 0) or
            not is_this_constant_value(self.y, 0) or
            not is_this_constant_value(self.width, 1920) or
            not is_this_constant_value(self.height, 1080) or
            not is_this_constant_value(self.rotation, 0)
        ):
            frame_processed = self._timeline_graph.render(
                inputs = {
                    'base_input': None,
                    'overlay_input': frame
                },
                evaluation_context = evaluation_context,
                output_size = self.output_size,
                do_use_gpu = self.do_use_gpu
            )

            # # The frame here can be 'ndarray' or 'Texture'
            # # depending on if the node was CPU or GPU
            return frame_processed

        # TODO: By now I'm not handling the 'opacity' as it is
        # not very important
        # if transform_state.opacity != 1:
        #     # TODO: Apply the opacity transformation
        #     pass

        return frame
    
    @staticmethod
    def default(
    ) -> 'VideoTransform':
        """
        Get an instance with the default values. Useful to
        instantiate it when `None` provided.
        """
        return VideoTransform(
            specification = ProgressSpecification(
                start_progress = 0.0,
                end_progress = 1.0
            ),
            output_size = VALUE_UNSET,
            # TODO: Maybe 'output_size = None' (?)
            # output_size = (1920, 1080),
            do_use_gpu = True,
            x = ConstantVideoEditorParameterValue(int(1920 / 2)),
            y = ConstantVideoEditorParameterValue(int(1080 / 2)),
            width = ConstantVideoEditorParameterValue(1920),
            height = ConstantVideoEditorParameterValue(1080),
            rotation = ConstantVideoEditorParameterValue(0),
            opacity = ConstantVideoEditorParameterValue(1.0)
        )
    
class AudioTransform(Transform):
    """
    The transformation conditions to define how we want
    to modify the different time moments of an audio in
    the timeline.

    This is attached to a timeline audio track item and
    will be applied on it. This is useful to apply the
    basic modifications.

    These values are time functions, and the values
    change according to the `t` time moment in which
    the values are requested: volume, intensity,
    equalization, noise reduction, etc.

    TODO: I think we should have 't_start' and 't_end'
    for another version that is applied on the track
    instead.
    """

    @property
    def copy(
        self
    ) -> 'AudioTransform':
        """
        Get a copy of it.
        """
        return AudioTransform(
            # TODO: Maybe we need a '.copy' of these below (?)
            volume = self.volume
        )

    def __init__(
        self,
        t_start: float,
        t_end: float,
        # TODO: By now I'm only modifying the volume
        volume: VideoEditorParameterValue = ConstantVideoEditorParameterValue(0.0)
    ):
        ParameterValidator.validate_positive_float('t_start', t_start, do_include_zero = True)
        ParameterValidator.validate_positive_float('t_end', t_end, do_include_zero = True)

        volume = parse_parameter(volume)

        # TODO: Still being developed
        self.volume: VideoEditorParameterValue = volume

    # def _get_state_at(
    #     self,
    #     t: float
    # ) -> AudioTransformState:
    #     """
    #     Get the `TransformState` dataclass that will indicate
    #     the values we have to apply to transform the timeline
    #     for the `t` time moment provided.
    #     """
    #     return AudioTransformState(
    #         volume = self.volume.get_value_at(t),
    #     )
    
    def apply(
        self,
        # TODO: Set the type
        frame: any,
        t: float
    # TODO: Set the type
    ) -> any:
        """
        Apply the transformation to the given `frame` at the also
        provided `t` (which will be used to obtain the transform
        state to apply).
        """
        transform_state = self._get_state_at(t)

        # TODO: Use the node and transform it directly
        # imitating the VideoTransform

        # Apply the effects to the 'frame' if needed
        if transform_state.volume != 1:
            # TODO: Apply the audio modifications
            # TODO: We don't have any audio modification node
            pass

        return frame
    
    # TODO: This should be updated when the audio is
    # being modified by a node with a TimelineGraph also
    def update(
        self,
        t_start: float,
        t_end: float
    ) -> 'VideoTransform':
        """
        Update the `t_start` and `t_end`.

        This method must be called each time the video
        duration this `AudioTransform` instance belongs
        to is modified.
        """
        self.t_start = t_start
        self.t_end = t_end

        return self
    
    @staticmethod
    def default(
    ) -> 'AudioTransform':
        """
        Get an instance with the default values. Useful to
        instantiate it when `None` provided.
        """
        return AudioTransform(
            t_start = 0.0,
            # TODO: I don't like this value
            t_end = 999.99,
            volume = ConstantVideoEditorParameterValue(0.0)
        )


from abc import ABC, abstractmethod


class Effect(ABC):
    """
    *Abstract class*

    Class to represent an effect that we will apply in
    a specific time moment, considering the params
    defined for it.

    An effect is something that uses a frame as an input
    and returns another input based on the changes, such
    as: blur, color correction, mask, chroma, etc.
    """

    @abstractmethod
    def _get_params_at(
        self,
        t: float
    ):
        """
        *For internal use only*

        Get the value of the parameters to apply at the given
        `t` time moment.    
        """
        pass

    @abstractmethod
    def apply(
        self,
        frame,
        t: float
    ):
        """
        Apply the effect to the given `frame` for the specific
        `t` time moment.
        """
        pass

class EffectParameters(ABC):
    """
    *Abstract class*

    Class to represent the parameters to use when
    applying an effect.
    """

    pass



from dataclasses import dataclass

@dataclass
class EffectInstance:
    """
    *Dataclass*

    Dataclass to include the information about the
    effect we want to apply.
    """

    def __init__(
        self,
        node_graph: 'NodeGraph',
        t_start: float,
        t_end: float,
        # blend_in: float = 0.0,
        # blend_out: float = 0.0
    ):
        self.node_graph: 'NodeGraph' = node_graph
        """
        The node that will be applied.
        """
        self.t_start: float = t_start
        """
        The time moment in which the effect must start
        being applied.
        """
        self.t_end: float = t_end
        """
        The time moment in which the effect must end
        being applied.
        """



"""
Our awesome editor module in which we
have all the classes that interact with
it and make it possible.

An editor includes a single timeline,
built by tracks, in which we place media
elements and we are able to apply effects
to them.

Here is a brief explanation about the
hierarchy:

- The editor has a timeline.
- The timeline has audio and video tracks
(that can handle when we play or not the
audio and video).
- The tracks have parts, that are virtual
items to simplify the way we combine 
tracks. Those parts include media 
instances. The tracks have priorities
ones against the others.
- The media instances have a time range
in which they must be played within the
track (within the timeline). Those media
instances have the start and end time
range for the media source and can apply
effects to the frames.
- The media sources are just the way we
obtain the frames for the specific media
items (read a file, constant color, etc.).
"""
from yta_editor.tracks.items.transition.transition_node_registry import register_transition_processor_nodes


def main():
    from yta_editor.tests import video_modified_stored

    video_modified_stored()


if __name__ == '__main__':
    pass
    # main()

# TODO: This should be somewhere else I think...
# Things I need to execute before launching the system
print('Launching things we need before...')
register_transition_processor_nodes()


from yta_editor.utils.frame_wrapper import AudioFrameWrapped, VideoFrameWrapped
from yta_editor.utils.frame_generator import VideoFrameGenerator
from yta_editor.transformations.transform import AudioTransform, VideoTransform, VALUE_UNSET
from yta_editor.transformations.effects.audio import AudioEffects
from yta_editor.transformations.effects.video import VideoEffects
from yta_editor.utils.transition_calculator import frame_to_uint8_ndarray
from yta_editor_parameters.values import ConstantVideoEditorParameterValue
from yta_editor_time.evaluation_context import EvaluationContext
from yta_logger import ConsolePrinter
from yta_video_frame_time.t_fraction import fps_to_time_base
from yta_time_interval import TimeInterval, TIME_INTERVAL_SYSTEM_LIMITS
from yta_validation.parameter import ParameterValidator
from quicktions import Fraction
from av import VideoFrame
from typing import Union
from abc import ABC, abstractmethod


# TODO: This is reapeated and not used everywhere
Number = Union[int, float, Fraction]
"""
Numeric type we accept as parameter.
"""

class _TrackItem(ABC):
    """
    Abstract class to represent an element that is on
    the track, that can be an empty space, a video, a
    transition or an audio.

    The time moments of this instance will be always a
    multiple of `1/fps`, using the `fps` of the track
    in which the item is located.
    
    This class must be inherited by our own custom item
    classes.
    """

    @property
    def t_start(
        self
    ) -> Fraction:
        """
        The time moment of the track in which this track item is
        placed, representing the moment in which it should start
        being played.
        """
        return self._time_interval.t_start

    @property
    def t_end(
        self
    ) -> Fraction:
        """
        The time moment of the track in which this track item is
        placed, representing the moment in which it should stop
        being played.
        """
        return self._time_interval.t_end
    
    @property
    def _t_first_frame(
        self
    ) -> Fraction:
        """
        *For internal use only*

        The time moment of the track in which the first frame of
        this track item starts being played, which is the `t_start`
        of this track item.

        The formula:
        - `self.t_start`
        """
        return self.t_start

    @property
    def _t_last_frame(
        self
    ) -> Fraction:
        """
        *For internal use only*

        The time moment of the track in which the last frame of
        this track item starts being played, which is the end of
        it minus 1/fps.

        The formula:
        - `self.t_end - (1 / self.fps)`
        """
        return self._get_frame_t_from_t_end(1)
    
    @property
    def duration(
        self
    ) -> Fraction:
        """
        The duration of this track item, based on its `t_start`
        and `t_end` time moments.

        The formula:
        - `self.t_end - self.t_start`
        """
        return self.t_end - self.t_start

    @property
    def fps(
        self
    ) -> float:
        """
        The fps of the track item, that is the same value as
        the track fps.
        """
        return self._track.fps
    
    @property
    def audio_fps(
        self
    ) -> float:
        """
        The audio fps of the track item, that is the same value
        as the track audio fps.
        """
        return self._track.audio_fps

    @property
    def t_start(
        self
    ) -> Fraction:
        """
        The global `t_start` time moment of this track item, which
        is obtained from the internal time interval.
        """
        return self._time_interval.t_start
    
    @property
    def t_end(
        self
    ) -> Fraction:
        """
        The global `t_end` time moment of this track item, which
        is obtained from the internal time interval.
        """
        return self._time_interval.t_end
    
    @property
    def duration(
        self
    ) -> Fraction:
        """
        The total duration of this track item based on its global
        `t_start` and `t_end` time moments.
        """
        return self._time_interval.duration
    
    @property
    @abstractmethod
    def copy(
        self
    ) -> '_TrackItem':
        """
        A copy of the current instance.
        """
        pass

    @property
    @abstractmethod
    def is_unchanged(
        self
    ) -> bool:
        """
        *For internal use only*

        Internal boolean flag to indicate if the source has
        been modified (or not), useful when we want to 
        optimize the way we access to the frames to render.

        If the original source has not been modified, we
        don't need to read frame by frame and apply any
        change so we can do it faster.

        (!) This property has to be modified by the specific
        audio sources.
        """
        pass

    def __init__(
        self,
        track: Union['AudioTrack', 'VideoTrack'],
        t_start: Union[int, float, Fraction],
        duration: Union[int, float, Fraction],
        item_in: Union['_AudioTrackItem', '_VideoTrackItem', 'TransitionTrackItem', 'GapTrackItem', None] = None,
        item_out: Union['_AudioTrackItem', '_VideoTrackItem', 'TransitionTrackItem', 'GapTrackItem', None] = None,
        transition_in: Union['TransitionTrackItem', None] = None,
        transition_out: Union['TransitionTrackItem', None] = None,
        # **kwargs
    ):
        """
        The `t_start` and `duration` values will be transformed into
        multiples of `1/fps`.
        """
        ParameterValidator.validate_mandatory_positive_number('t_start', t_start, do_include_zero = True)
        ParameterValidator.validate_mandatory_positive_number('duration', duration, do_include_zero = False)

        self._track: Union['AudioTrack', 'VideoTrack'] = track
        """
        The instance of the track this track item belongs to.
        """

        self._time_interval: TimeInterval = TimeInterval(
            t_start = t_start,
            t_end = t_start + duration,
            t_start_limit = None,
            t_end_limit = None,
            # TODO: This 'duration' has to be used when media
            duration_limit = None,
            fps = self.fps
        )
        """
        *For internal use only*

        Internal time interval instance to be able to manage the
        `t_start` and `t_end` time moments properly.
        """

        self.item_in: Union['_AudioTrackItem', '_VideoTrackItem', 'TransitionTrackItem', 'GapTrackItem', None] = None
        """
        The previous track item.
        """
        self.item_out: Union['_AudioTrackItem', '_VideoTrackItem', 'TransitionTrackItem', 'GapTrackItem', None] = None
        """
        The next track item.
        """
        self.transition_in: Union['TransitionTrackItem'] = None
        """
        The transition that will make the previous item in the
        track be mixed with this one at the begining.
        """
        self.transition_out: Union['TransitionTrackItem'] = None
        """
        The transition that will make the next item in the
        track be mixed with this one at the t_end.
        """

        # We will set it like this to force recalculations if needed
        if item_in is not None:
            self.set_item_in(item_in)
        if item_out is not None:
            self.set_item_out(item_out)
        if transition_in is not None:
            self.set_transition_in(transition_in)
        if transition_out is not None:
            self.set_transition_out(transition_out)

    def set_item_in(
        self,
        item: Union['_AudioTrackItem', '_VideoTrackItem', 'TransitionTrackItem', 'GapTrackItem'],
        do_propagate: bool = True,
    ) -> '_TrackItem':
        """
        Set the provided `item` as the previous one (`.item_in`),
        and this one (self) as the `.item_out` of the one provided
        as parameter, only if the `do_propagate` parameter is True.
        """
        # TODO: Validate as one of those instances and not None
        self.item_in = item

        if (
            item.item_out is not self and
            do_propagate
        ):
            item.set_item_out(
                item = self,
                do_propagate = False
            )
        # TODO: Update the others affected if necessary (?)

        return self
    
    def unset_item_in(
        self,
        do_propagate: bool = True
    ) -> '_TrackItem':
        """
        Unset the previous item (`.item_in`) from this instance
        but also unset the `.item_out` of the previous one only if
        the `do_propagate` parameter is True.
        """
        if self.item_in is not None:
            item = self.item_in
            self.item_in = None
            # TODO: Setting 'None' maybe should be setting a Gap...

            if (
                item.item_out is self and
                do_propagate
            ):
                item.unset_item_out(do_propagate = False)

        return self
    
    def set_item_out(
        self,
        item: Union['_AudioTrackItem', '_VideoTrackItem', 'TransitionTrackItem', 'GapTrackItem'],
        do_propagate: bool = True
    ) -> '_TrackItem':
        """
        Set the provided `item` as the next one (`.item_out`),
        and this item (self) as the `.item_in` of the `item`
        provided only if the `do_propagate` parameter is True.
        """
        self.item_out = item

        if (
            item.item_in is not self and
            do_propagate
        ):
            item.set_item_in(
                item = self,
                do_propagate = False
            )
        # TODO: Update the others affected if necessary (?)
        
        return self
    
    def unset_item_out(
        self,
        do_propagate: bool = True
    ) -> '_TrackItem':
        """
        Unset the previous item (`.item_out`) from this instance
        but also unset the `.item_in` of the previous one only if
        the `do_propagate` parameter is True.
        """
        if self.item_out is not None:
            item = self.item_out
            self.item_out = None
            # TODO: Setting 'None' maybe should be setting a Gap...

            if (
                item.item_in is self and
                do_propagate
            ):
                item.unset_item_in(do_propagate = False)

        return self
    
    def set_transition_in(
        self,
        transition: 'TransitionTrackItem'
    ) -> '_TrackItem':
        """
        Set the `transition` provided as the transition to be applied
        to mix the previous item `.item_in` with this one.
        
        TODO: Do the 'decrease' here or in the 'in' (?)

        This method will update the `t_start` time of the next items 
        (decreasing it) according to the duration of the transition
        applied, as the next items will appear earlier.
        """
        if self.item_in is None:
            raise Exception('There is no `.item_in` (previous item) to transition with.')
        
        self.transition_in = transition
        # TODO: We need to update (decrease 'self.duration') the next
        # items starts

        return self
    
    def unset_transition_in(
        self
    ) -> '_TrackItem':
        """
        Unset the transition with the previous item (`.transition_in`).

        TODO: Do the 'reset' here or in the 'in' (?)

        This method will update the `t_start` time of the next items 
        (increasing it) according to the duration of the transition
        applied, as the next items will appear later.
        """
        self.transition_in = None
        # TODO: We need to update (increase 'self.duration') the next
        # items starts

        return self
    
    def set_transition_out(
        self,
        transition: 'TransitionTrackItem'
    ) -> '_TrackItem':
        """
        Set the `transition` provided as the transition to be applied
        to mix the next item `.item_out` with this one.
        
        TODO: Do the 'decrease' here or in the 'in' (?)
        
        This method will update the `t_start` time of the next items 
        (decreasing it) according to the duration of the transition
        applied, as the next items will appear earlier.
        """
        if self.item_out is None:
            raise Exception('There is no `.item_out` (next item) to transition with.')
        
        self.transition_out = transition
        # TODO: The 'decrease' will be done here or in the 'set_transition_in' (?)

        return self
    
    def unset_transition_out(
        self
    ) -> '_TrackItem':
        """
        Unset the transition with the next item (`.transition_out`).

        TODO: Do the 'reset' here or in the 'in' (?)

        This method will update the `t_start` time of the next items 
        (increasing it) according to the duration of the transition
        applied, as the next items will appear later.
        """
        self.transition_out = None
        # TODO: The 'decrease' will be done here or in the 'set_transition_in' (?)

        return self
    
    """
    Functionality related to the time interval below.
    """

    def _shift_by(
        self,
        delta: Number
    ) -> '_TrackItem':
        """
        *For internal use only*

        This method should be called by the `Track` instance that
        holds this item.

        Move the `t_start` and `t_end` value of this item the amount
        of time defined by the `delta` provided parameter, that
        can be positive or negative (and will be truncated to be
        a multiple of `1/fps`).

        The `t_end` value will never go beyond the system's limit,
        so the duration of the item will be enshortened when delta
        is too high.
        """
        if (self.t_end + delta) >= TIME_INTERVAL_SYSTEM_LIMITS[1]:
            self._time_interval.with_start_and_end(
                t_start = self.t_start + delta,
                t_end = TIME_INTERVAL_SYSTEM_LIMITS[1]
            )
        else:
            self._time_interval.shift_by(delta)

        return self
    
    # TODO: Separate the other methods as '_shift_by' and 'shift_by'
    def shift_to(
        self,
        t_track: Number
    ) -> '_TrackItem':
        """
        Move the `t_start` and `t_end` value of this item to the time
        moment defined by the `t_track` parameter provided. The
        change will not modify the duration of the item.
        """
        return self._shift_by(
            delta = t_track - self.t_start
        )
    
    def shift_start_by(
        self,
        delta: Number
    ) -> '_TrackItem':
        """
        Update the `t_start` value by adding the `delta` provided
        (that can be positive to make it start later or negative
        to make it start earlier).
        """
        self._time_interval.shift_start_by(delta)

        return self
    
    def shift_start_to(
        self,
        t_track: Number
    ) -> '_TrackItem':
        """
        Update the `t_start` value to the one provided as `t_track`
        (that can be before or after the current one).
        """
        return self.shift_start_by(
            delta = t_track - self.t_start
        )
    
    def shift_end_by(
        self,
        delta: Number
    ) -> '_TrackItem':
        """
        Update the `t_end` value by adding the `delta` provided
        (that can be positive to make it start later or negative
        to make it start earlier).
        """
        self._time_interval.shift_end_by(delta)

        return self
    
    def shift_end_to(
        self,
        t_track: Number
    ) -> '_TrackItem':
        """
        Update the `t_end` value to the one provided as `t_track`
        (that can be before or after the current one).
        """
        return self.shift_end_by(
            delta = t_track - self.t_end
        )
    
    def split(
        self,
        t_track: Number
    ) -> '_TrackItem':
        """
        Split the current item at the `t_track` track time
        moment provided and get the 2 new TimeInterval
        instances as a tuple:
        - A: `[t_start, t_track)`
        - B: `[t_track, t_end)`.
        """
        return self._time_interval.split(t_track)
    
    def cut(
        self,
        track_start: Number,
        track_end: Number
    ) -> '_TrackItem':
        """
        Transform this track item into a new one which
        `track_start` and `track_end` are now the ones
        provided as parameters.
        """
        # TODO: What if 't_start' and 't_end' are greater than
        # before? It is not actually cutting...
        self._time_interval.cut(track_start, track_end)

        return self
    
    # TODO: What (?)
    #def propagate_
    
    """
    Functionality related to the time interval above.
    """
    
    def __str__(
        self
    ) -> str:
        """
        Function to stringify the instance and show the in
        and out items next to the item itself.
        """
        item_in: str = (
            f'{self.item_in.__class__.__name__} [{str(float(self.item_in.t_start))}, {str(float(self.item_in.t_end))}]'
            if self.item_in is not None else
            '--'
        )

        item_out: str = (
            f'{self.item_out.__class__.__name__} [{str(float(self.item_out.t_start))}, {str(float(self.item_out.t_end))}]'
            if self.item_out is not None else
            '--'
        )

        return f'  > (In: {item_in}) -   | {self.__class__.__name__} [{str(float(self.t_start))}, {str(float(self.t_end))}] |   - (Out: {item_out})'
    
    def _get_frame_t_from_t_start(
        self,
        number_of_frames: int
    ):
        """
        *For internal use only*

        Get the exact time moment of the frame that is 'number_of_frames'
        frames ahead of the 't_start' time moment of this track item.
        """
        ParameterValidator.validate_mandatory_positive_int('number_of_frames', number_of_frames, True)

        return self._time_interval._t_handler.t.next(self.t_start, number_of_frames) 

    def _get_frame_t_from_t_end(
        self,
        number_of_frames: int
    ):
        """
        *For internal use only*

        Get the exact time moment of the frame that is 'number_of_frames'
        frames before the 't_end' time moment of this track item.
        """
        ParameterValidator.validate_mandatory_positive_int('number_of_frames', number_of_frames, True)

        return self._time_interval._t_handler.t.previous(self.t_end, number_of_frames)
    
    def _t_track_to_t_track_item(
        self,
        # TODO: Use 'TimeMoment'
        t_track: float
    # TODO: Use 'TimeMoment'
    ) -> float:
        """
        *For internal use only*

        Get the local `t` time moment of this track item
        according to the `t_track` provided.

        The formula:
        - `t_track - self.t_start`
        """
        return t_track - self.t_start
    
    def _get_evaluation_context(
        self,
        t_track: float
    ) -> 'EvaluationContext':
        """
        *For internal use only*

        Get the evaluation context for this track item.
        The evaluation context must be passed to the
        effects and other elements that need to make
        calculations based on the frame index, time
        moment, etc.
        """
        return EvaluationContext.from_t(
            fps = self.fps,
            total_frames = int(self.duration * self.fps),
            t = self._t_track_to_t_track_item(t_track)
        )
    
        return get_evaluation_context(
            t_track = t_track,
            t_start_item = self.t_start,
            t_end_item = self.t_end,
            fps = self.fps,
            track_duration = self._track.t_end
        )
    
class _TrackItemWithMedia(_TrackItem):
    """
    Abstract class to represent a track item that includes a
    media file to obtain the frames from. This item is 
    special because its time interval and cannot be shifted
    randomly.

    Here you have some specific cases:
    - If the item is shifted in general, only the item time
    interval will be moved, the media time interval will be
    the same.
    - If the start of the item is shifted, the media's start
    will be also shifted to fit the same duration (this is
    like enshorting or enlarging a video from the left).
    - If the end of the item is shifted, the media's end will
    be also shifted to fit the same duration (this is like
    enshorting or enlarging a video from the right).

    Based on this before, if you need an item to

    """

    @property
    def start_limit(
        self
    ) -> Fraction:
        """
        The global time moment that is the minimum `t` time moment
        this item should be able to handle according to the media
        and the source related to this media.

        Considering a source of 8s, if this item is placed in the
        [9, 11) of the track, it should accept any `t` time moment
        in between those 9 and 11 limits. But, as the media associated
        could be, for example, [1, 3), a `t_track=8` could be accepted
        in special cases (transitions) because it would be transformed
        into `t_media=-1`, which would return `t_source=0` that is
        valid. But not a `t_track=7`, because it would be `t_media=-2`,
        which is `t_source=-1` (invalid).

        This `start_limit` must be used carefully and only for
        special cases (such as transitions).
        """
        # The 'self.media.start_limit' will be always negative or 0
        return self.t_start + self.media.start_limit

    @property
    def end_limit(
        self
    ) -> Fraction:
        """
        The global time moment that is the maximum `t` time moment
        this item should be able to handle according to the media
        and the source related to this media.
        
        Considering a source of 8s, if this item is placed in the
        [9, 11) of the track, it should accept any `t` time moment
        in between those 9 and 11 limits. But, as the media associated
        could be, for example, [2, 4), a `t_track=12` could be accepted
        in special cases (transitions) because it would be transformed
        into `t_media=3`, which would return `t_source=5` that is
        valid. But not a `t_track=16`, because it would be `t_media=7`,
        which is `t_source=9` (invalid).

        This `end_limit` must be used carefully and only for
        special cases (such as transitions).
        """
        return self.t_start + self.media.end_limit
    
    @property
    def is_unchanged(
        self
    ) -> bool:
        """
        *For internal use only*

        Internal boolean flag to indicate if the source has
        been modified (or not), useful when we want to 
        optimize the way we access to the frames to render.

        If the original source has not been modified, we
        don't need to read frame by frame and apply any
        change so we can do it faster.
        """
        return self.media.is_unchanged

    def __init__(
        self,
        t_start: Number,
        media: Union['AudioMedia', 'VideoMedia'],
        **kwargs
    ):
        # TODO: I think I don't need 'AudioTime' nor 'VideoTimed'
        # but only 'AudioMedia' and 'VideoMedia'
        # ParameterValidator.validate_mandatory_instance_of('media', media, ['AudioMedia', 'VideoMedia'])

        # TODO: I think I don't need 'AudioTime' nor 'VideoTimed'
        # but only 'AudioMedia' and 'VideoMedia'
        self.media: Union['AudioMedia', 'VideoMedia'] = media
        """
        The media associated to this track item.
        """

        super().__init__(
            t_start = t_start,
            duration = media.duration,
            **kwargs
        )
        
        self._time_interval.duration_limit = self.media.max_duration
        
    def _t_track_to_t_media(
        self,
        t_track: Union[int, float, Fraction],
        do_use_source_limits: bool = False
    ) -> Fraction:
        """
        *For internal use only*

        Get the local `t` time moment value for the media
        associated, based on the global `t` time moment
        provided.

        This method will raise an exception if the `t`
        time moment provided is not included in this track
        item time interval and the `do_use_source_limits`
        parameter is not `True`. Send it as `True` if you
        need to force reading the media maybe for transitions.

        The formula:
        - `t_track - self.t_start`
        """
        start_limit = (
            self.start_limit
            if do_use_source_limits else
            self._time_interval.t_start
        )

        end_limit = (
            self.end_limit
            if do_use_source_limits else
            self._time_interval.t_end
        )

        ParameterValidator.validate_mandatory_number_between('t_track', t_track, start_limit, end_limit, True, False)
        
        """
        Imagine that we have this item in [4, 8] and we
        receive t=6. The local_t=6-4=2. The media we have
        available has 4s of duration (so [0, 4]), but it
        is actually cropped to [1, 3], so the moment we
        need to read is the t=2 of the cropped version,
        so: t_media=1+2=3. But is the media who will
        calculate that internal t, we only need to ask him
        for the t=2 and he will make the conversion (that
        is currently being done by a decorator).

        That means that we could actually ask for the t=-1
        to the media, that would be turned into the t=0 of
        the source which is actually available, so we have
        to be very careful with this...

        Source [0, 8)
        Media [0, 2) but of the source [2, 4)
        TrackItem [11, 13)
        -> Ask t_t=12 => t_media=1 => t_source = 3
            Is valid and included in the normal limits
        -> Ask t_t=10 => t_media=-1 => t_source=1
            Is valid only if forcing to use the source
            limits because t_source=1, but t_t=10 is out
            of the [11, 13) time interval
        """

        # TODO: Do we keep this new option (?)
        return super()._t_track_to_t_track_item(t_track)
        return t_track - self.t_start
    
    # The super() methods will shift only the item
    """
    These methods below will shift the media only.
    """
    def shift_media_by(
        self,
        delta: Union[int, float, Fraction]
    ) -> '_TrackItemWithMedia':
        """
        Update the media's `t_start` and `t_end` time moments by
        the `delta` amount of time provided.

        I'll give you one example below:
        - We have the item at [1, 4), and the media is pointing to
        the [2, 5) of the source, and we do `.shift_media_by(1)`.
        The item will be still [1, 4), but the media will be
        pointing to the [3, 6) time interval of the source.
        """
        self.media.shift_by(delta)
    
        return self
    
    def shift_media_start_and_item_end_by(
        self,
        delta: Union[int, float, Fraction]
    ) -> '_TrackItemWithMedia':
        """
        Update the media's `t_start` time moment by the `delta`
        amount of time provided, but the `t_end` of the media is
        not affected, and updates the item `t_end` time moment by
        that same `delta` amount of time, affecting not to the
        item's `t_start` time moment.

        I'll give you one example below:
        - We have the item at [1, 4), and the media is pointing to
        the [2, 5) of the source, and we do
        `.shift_media_start_and_item_end_by(1)`. The item will be
        [1, 3), but the media will be pointing to the [4, 6) time
        interval of the source.
        """
        self.media.shift_start_by(delta)

        self._time_interval.shift_end_by(-delta)

        return self

    def shift_media_to(
        self,
        t_media: Union[int, float, Fraction]
    ) -> '_TrackItemWithMedia':
        """
        Update the media's `t_start` to the `t_media` time moment
        provided, and the `t_end` the same amount of time.

        I'll give you one example below:
        - We have the item at [1, 4), and the media is pointing to
        the [2, 5) of the source, and we do `.shift_media_to(3)`.
        The item will be still [1, 4), but the media will be
        pointing to the [3, 6) time interval of the source.
        """
        delta = t_media - self.media.t_start

        return self.shift_media_by(delta)

    """
    These methods below will shift both the item and the media.
    """
    def shift_item_and_media_by(
        self,
        delta: Union[int, float, Fraction]
    ) -> '_TrackItemWithMedia':
        """
        TODO: Shifting item and media at the same time? I don't
        think so... (?)

        Update the item `t_start` time moment by the `delta` value
        provided but also the media's `t_start` by the same `delta`
        value, and the `t_end` time moment of both will be also
        modified with the same amount of time. This value can be
        positive (will make it start later) or negative (will make
        it start earlier).

        I'll give you some examples below:
        - We have the item at [1, 4), and the media is pointing to
        the [2, 5) of the source, and we do
        `.shift_item_and_media_by(1)`. The item will be now [2, 5),
        and the media will be pointing to the [3, 6) time interval
        of the source.
        - We have the item at [1, 4), and the media is pointing to
        the [2, 5) of the media, and we do
        `.shift_item_and_media_by(-1)`. The item will be now [0, 3),
        and the media will be pointing to the [1, 4) time interval
        of the source.
        - We have the item at [3, 6), and the media is pointing to
        the [2, 5) of the source, and we do
        `.shift_item_and_media_by(-3)`. The item can be [0, 6)
        because the duration of the media is 8s, but the media would
        be pointing to the source's [-1, 2) time interval, which is
        not possible because is out of the limits (minimum `t_start
        value is 0)..
        
        This is limited, at first, by the item time interval limit
        but also by the media time interval limit.
        """
        # TODO: Validate that the new `start` is valid according
        # to the item time interval limits
        # TODO: Validate that the new `media_start` is valid
        # according to the media interval limit
        # TODO: Is this above validated when 'shifting' (?)
        self.media.shift_by(delta)

        self._shift_by(delta)

        return self

    def shift_item_and_media_to(
        self,
        t_track: Union[int, float, Fraction]
    ) -> '_TrackItemWithMedia':
        """
        TODO: Shifting item and media at the same time? I don't
        think so... (?)

        Update the item `t_start` time moment to the `t_track` time
        moment provided but also the media's `t_start` time 
        variation, and the `t_end` time moment of both will be also
        modified with the same amount of time. This value can be
        greater (will make it start later) or lower (will make
        it start earlier) than the current track `t_start` time
        moment.

        I'll give you some examples below:
        - We have the item at [1, 4), and the media is pointing to
        the [2, 5) of the source, and we do
        `.shift_item_and_media_to(2)`. The item will be now [2, 5),
        and the media will be pointing to the [3, 6) time interval
        of the source.
        - We have the item at [1, 4), and the media is pointing to
        the [2, 5) of the media, and we do
        `.shift_item_and_media_to(0)`. The item will be now [0, 3),
        and the media will be pointing to the [1, 4) time interval
        of the source.
        - We have the item at [3, 6), and the media is pointing to
        the [2, 5) of the source, and we do
        `.shift_item_and_media_to(0)`. The item can be [0, 6)
        because the duration of the media is 8s, but the media would
        be pointing to the source's [-1, 2) time interval, which is
        not possible because is out of the limits (minimum `t_start
        value is 0)..
        
        This is limited, at first, by the item time interval limit
        but also by the media time interval limit.
        """
        delta = t_track - self.t_start
        
        return self.shift_item_and_media_by(delta)

    """
    These methods below will shift the start and the end
    of the item and the media at the same time.
    """
    def shift_start_by(
        self,
        delta: Union[int, float, Fraction]
    ) -> '_TrackItemWithMedia':
        """
        Update the item `t_start` time moment by the `delta` value
        provided but also the media's `t_start` by the same `delta`
        value. This value can be positive (will make it last less
        time) or negative (will make it longer).

        I'll give you some examples below:
        - We have the item at [1, 4), and the media is pointing to
        the [2, 5) of the source, and we do `.shift_start_by(1)`.
        The item will be now [2, 4), and the media will be pointing
        to the [3, 5) time interval of the source.
        - We have the item at [1, 4), and the media is pointing to
        the [2, 5) of the media, and we do `.shift_start_by(-1)`.
        The item will be now [0, 4), and the media will be pointing
        to the [1, 5) time interval of the source.
        - We have the item at [3, 6), and the media is pointing to
        the [2, 5) of the source, and we do `.shift_start_by(-3)`.
        The item can be [0, 6) because the duration of the media is
        8s, but the media would be pointing to the source's [-1, 2)
        time interval, which is not possible because is out of the
        limits (minimum `t_start` value is 0)..
        
        This is limited, at first, by the item time interval limit
        but also by the media time interval limit.
        """
        # TODO: Validate that the new `start` is valid according
        # to the item time interval limits
        # TODO: Validate that the new `media_start` is valid
        # according to the media interval limit
        # TODO: Is this above validated when 'shifting' (?)
        self.media.shift_start_by(delta)

        super().shift_start_by(delta)

        return self
    
    def shift_start_to(
        self,
        t_track: Union[int, float, Fraction]
    ) -> '_TrackItemWithMedia':
        """
        Update the item `t_start` time moment to the `t_track` time
        moment value provided, but also the media the same amount
        of time than the item (only if possible). Be careful with
        the limits of the media's time interval.

        I'll give you some examples below with a media with 8s of
        duration:
        - We have the item at [1, 4), and the media is pointing to
        the [2, 5) of the source, and we do `.shift_start_to(3)`.
        The item will be now [3, 4), and the media will be pointing
        to the [4, 5) of the source.
        - We have the item at [1, 4), and the media is pointing to
        the [2, 5) of the source, and we do `.shift_start_to(0)`.
        The item will be now [0, 4), and the media will be pointing
        to the [1, 5) of the source.
        - We have the item at [3, 6), and the media is pointing to
        the [2, 5) of the source, and we do `.shift_start_to(0)`.
        The item can be [0, 6) because the duration of the media is
        8s, but the media would be pointing to the source's [-1, 2)
        time interval, which is not possible because is out of the
        limits (minimum `t_start` value is 0).
        
        This is limited, at first, by the item time interval limit
        but also by the media time interval limit.
        """
        delta = t_track - self.t_start

        # This will force the 2 shifts according to their limits
        return self.shift_start_by(delta)
    
    def shift_end_by(
        self,
        delta: Union[int, float, Fraction]
    ) -> '_TrackItemWithMedia':
        """
        Update the item `t_end` time moment by the `delta` value
        provided but also the media's `t_end` by the same `delta`
        value. This value can be positive (will make it last more
        time) or negative (will make it shorter).

        I'll give you some examples below:
        - We have the item at [1, 4), and the media is pointing to
        the [2, 5) of the source, and we do `.shift_end_by(1)`.
        The item will be now [1, 5), and the media will be pointing
        to the [2, 6) time interval of the source.
        - We have the item at [1, 4), and the media is pointing to
        the [2, 5) of the media, and we do `.shift_end_by(-1)`.
        The item will be now [1, 3), and the media will be pointing
        to the [2, 4) time interval of the source.
        - We have the item at [3, 6), and the media is pointing to
        the [2, 5) of the source, and we do `.shift_end_by(4)`.
        The item can be [3, 10) because the duration of the media is
        8s, but the media would be pointing to the source's [2, 9)
        time interval, which is not possible because is out of the
        limits (maximum `t_end` value is 8).
        
        This is limited, at first, by the item time interval limit
        but also by the media time interval limit.
        """
        # TODO: Validate that the new `start` is valid according
        # to the item time interval limits
        # TODO: Validate that the new `media_start` is valid
        # according to the media interval limit
        # TODO: Is this above validated when 'shifting' (?)
        self.media.shift_end_by(delta)

        super().shift_end_by(delta)

        return self
    
    def shift_end_to(
        self,
        t_track: Union[int, float, Fraction]
    ) -> '_TrackItemWithMedia':
        """
        Update the item `t_end` time moment to the `t_track` time
        moment value provided, but also the media the same amount
        of time than the item (only if possible). Be careful with
        the limits of the media's time interval.

        I'll give you some examples below:
        - We have the item at [1, 4), and the media is pointing to
        the [2, 5) of the source, and we do `.shift_end_to(5)`.
        The item will be now [1, 5), and the media will be pointing
        to the [2, 6) time interval of the source.
        - We have the item at [1, 4), and the media is pointing to
        the [2, 5) of the media, and we do `.shift_end_to(3)`.
        The item will be now [1, 3), and the media will be pointing
        to the [2, 4) time interval of the source.
        - We have the item at [3, 6), and the media is pointing to
        the [2, 5) of the source, and we do `.shift_end_to(10)`.
        The item can be [3, 10) because the duration of the media is
        8s, but the media would be pointing to the source's [2, 9)
        time interval, which is not possible because is out of the
        limits (maximum `t_end` value is 8).
        
        This is limited, at first, by the item time interval limit
        but also by the media time interval limit.
        """
        delta = t_track - self.t_end

        # This will force the 2 shifts according to their limits
        return self.shift_end_by(delta)
    
    # TODO: We need these methods:
    # - shift_only_item: moves the item time interval but not media
    # - shift: moves the item time interval but also the media
    # - shift_only_media: moves the media time interval only
    # Any other method (shift_start, shift_end) will shift both
    # because it is necessary to do it

class _TrackItemWithAudio(_TrackItem):
    """
    Abstract class to implement a track item that is capable
    of providing audio frames.
    """

    @property
    def _first_audio_frames(
        self
        # TODO: What about the type (?)
    ):
        """
        *For internal use only*
        
        The audio frames of the first video frame (as a generator),
        to be used when the item is in a transition in which the
        first and last frames are frozen to not affect the duration.

        This is a cached property that will change if the
        corresponding time moment is changed because the item gets
        trimmed or extended.
        """
        if (
            not hasattr(self, '_first_audio_frames_cached') or
            self._t_first_audio_frames_cached != self._t_first_frame
        ):
            self._t_first_audio_frames_cached = self._t_first_frame
            self._first_audio_frames_cached = self.get_audio_frames_at(
                t_track = self._t_first_audio_frames_cached,
                video_fps = self._track.fps,
                do_use_source_limits = False
            )

        return self._first_audio_frames_cached
    
    @property
    def _last_audio_frames(
        self
        # TODO: What about the type (?)
    ):
        """
        *For internal use only*
        
        The audio frames of the last video frame (as a generator),
        to be used when the item is in a transition in which the
        first and last frames are frozen to not affect the duration.

        This is a cached property that will change if the
        corresponding time moment is changed because the item gets
        trimmed or extended.
        """
        if (
            not hasattr(self, '_last_audio_frames_cached') or
            self._t_last_audio_frames_cached != self._t_last_frame
        ):
            self._t_last_audio_frames_cached = self._t_last_frame
            self._last_audio_frames_cached = self.get_audio_frames_at(
                t_track = self._t_last_audio_frames_cached,
                video_fps = self._track.fps,
                do_use_source_limits = False
            )

        return self._last_audio_frames_cached
    
    def __init__(
        self,
        audio_transform: Union[AudioTransform, None] = None,
        audio_effects: Union[AudioEffects, None] = None,
        **kwargs
    ):
        # TODO: What about the time moments (?)
        self.audio_transform: Union[AudioTransform, None] = audio_transform
        """
        The transformations we want to apply in the audio.
        """
        self.audio_effects: Union[AudioEffects, None] = audio_effects
        """
        The effects we want to apply in the audio.
        """

        super().__init__(
            **kwargs
        )

    def get_audio_frames_at(
        self,
        t_track: Union[int, float, Fraction],
        # TODO: Ignore this, we have the 'self.fps' here
        video_fps: Union[int, float, Fraction] = None,
        do_use_source_limits: bool = False
    ):
        """
        Iterate over all the audio frames that
        exist at the time moment 't' provided.
        """
        # frames = []
        # # TODO: This should be different with the new gap item
        # if not self.is_gap:
        #     # TODO: What do we do in this case (?)
        #     frames = list(self._get_audio_frames_at(t_track))

        #     if len(frames) == 0:
        #         ConsolePrinter().print(f'   [ERROR] Audio frame {str(float(t_track))} was not obtained')
        #     else:
        #         frames = [
        #             AudioFrameWrapped(
        #                 frame = frame,
        #                 is_from_gap = False
        #             )
        #             for frame in frames
        #         ]
        # TODO: Remove this on top if below is working

        frames = list(self._get_audio_frames_at(
            t_track = t_track,
            video_fps = video_fps,
            do_use_source_limits = do_use_source_limits
        ))

        if len(frames) == 0:
            ConsolePrinter().print(f'   [ERROR] Audio frame {str(float(t_track))} was not obtained')
        else:
            # TODO: We should not receive AudioFrame and 
            # AudioFrameWrapped here, only one. Fix it (!)
            from yta_validation import PythonValidator
            frames = [
                (
                    AudioFrameWrapped(
                        frame = frame,
                        is_from_gap = False
                    )
                )
                if PythonValidator.is_instance_of(frame, 'AudioFrame') else
                frame
                for frame in frames
            ]

        frames = (
            # TODO: We couldn't obtain the frames (?)
            self._track.audio_silent
            if len(frames) == 0 else
            frames
        )

        for frame in frames:
            frame = (
                self.audio_transform.apply(frame, t_track)
                if self.audio_transform is not None else
                frame
            )

            yield frame

    @abstractmethod
    def _get_audio_frames_at(
        self,
        t_track: Union[int, float, Fraction],
        # TODO: Ignore this, we have 'self.fps' here
        video_fps: Union[int, float, Fraction] = None,
        do_use_source_limits: bool = False
    ):
        """
        Get the audio frames of the provided `t` global time
        moment according to the internal configuration of 
        the track item.

        TODO: This method must be overwritten by the
        specific class implementations.
        """
        pass

class TrackItemWithAudioMedia(_TrackItemWithAudio, _TrackItemWithMedia):
    """
    A track item that is capable of providing audio
    frames by obtaining them from the media it contains.
    """

    # TODO:I think we need the 'copy' as in 'TrackItemWithVideoMedia'

    def __init__(
        self,
        track: 'AudioTrack',
        t_start: Union[int, float, Fraction],
        media: 'AudioMedia',
        audio_transform: Union[AudioTransform, None] = None,
        audio_effects: Union['AudioEffects', None] = None,
        item_in: Union['_VideoTrackItem', 'TransitionTrackItem', 'GapTrackItem', None] = None,
        item_out: Union['_VideoTrackItem', 'TransitionTrackItem', 'GapTrackItem', None] = None,
        transition_in: Union['TransitionTrackItem', None] = None,
        transition_out: Union['TransitionTrackItem', None] = None
    ):
        super().__init__(
            track = track,
            t_start = t_start,
            media = media,
            audio_transform = audio_transform,
            audio_effects = audio_effects,
            item_in = item_in,
            item_out = item_out,
            transition_in = transition_in,
            transition_out = transition_out
        )

    # TODO: Why do I have 2 different '__init__' (?)
    def __init__(
        self,
        media: 'AudioTimed',
        audio_transform: Union[AudioTransform, None] = None,
        audio_effects: Union[AudioEffects, None] = None,
        # TODO: Should we use the parameters that are in the other
        # '__init__()' method
        **kwargs
    ):
        ParameterValidator.validate_mandatory_instance_of('media', media, 'AudioTimed')

        super().__init__(
            media = media,
            audio_transform = audio_transform,
            audio_effects = audio_effects,
            **kwargs
        )

        self._synchronize_audio_transform()

    def _synchronize_audio_transform(
        self
    ) -> 'TrackItemWithVideoMedia':
        """
        *For internal use only*

        Synchronize the audio transform instance by
        setting the `t_start` and `t_end` according to
        this track item time moments.
        """
        self.audio_transform = (
            self.audio_transform.update(
                t_start = self.t_start,
                t_end = self.t_end
            )
            if self.audio_transform is not None else
            None
        )

        return self

    def _get_audio_frames_at(
        self,
        t_track: Union[int, float, Fraction],
        # TODO: Ignore this, we have 'self.fps' here
        video_fps: Union[int, float, Fraction] = None,
        do_use_source_limits: bool = False
    ):
        """
        Get the audio frames for the provided `t` global
        time moment, that is transformed into the media
        local time moment, from the audio media this
        instance contains.

        The `do_use_source_limits` method should be used
        when we want to access to a `t_track` that is out
        of the time interval of this track item but it is
        actually accepted by the internal source time
        interval limits.
        """
        t_media = self._t_track_to_t_media(
            t_track = t_track,
            do_use_source_limits = do_use_source_limits
        )

        return self.media.get_audio_frames_at(
            t_media = t_media,
            video_fps = self.fps,
            do_use_source_limits = do_use_source_limits
        )
    
class _TrackItemWithVideo(_TrackItem):
    """
    Abstract class to implement a track item that is capable
    of providing audio frames.
    """

    @property
    def _frame_not_found(
        self
    ):
        """
        *For internal use only*

        The frame we want to send to the editor when
        we were not able to find the frame we were
        looking for (due to an error or that the file
        was unreadable in that specific moment).

        This property has been created to avoid
        creating these empty frames again and again,
        as they will be always the same.

        TODO: This should not happen, it is an
        internal way to point an error, and of
        course should be removed in a future 
        version when we don't suffer it.
        """
        if not hasattr(self, '__frame_not_found'):
            self.__frame_not_found = VideoFrameGenerator.full_red(
                size = self._track.output_size,
                time_base = fps_to_time_base(self._track.fps)
            )

        return self.__frame_not_found
    
    @property
    def _first_video_frame(
        self
        # TODO: What about the type (?)
    ):
        """
        *For internal use only*

        The first video frame of this track item, useful when using
        the item in a transition in which the first and the last
        frames are frozen to not affect the duration.

        This is a cached property that will change if the
        corresponding time moment is changed because the item gets
        trimmed or extended.
        """
        if (
            not hasattr(self, '_first_video_frame_cached') or
            self._t_first_video_frame_cached != self._t_first_frame
        ):
            self._t_first_video_frame_cached = self._t_first_frame
            self._first_video_frame_cached = self.get_video_frame_at(
                t_track = self._t_first_video_frame_cached,
                do_use_source_limits = False
            )

        return self._first_video_frame_cached
    
    @property
    def _last_video_frame(
        self
        # TODO: What about the type (?)
    ):
        """
        *For internal use only*

        The last video frame of this track item, useful when using
        the item in a transition in which the first and the last
        frames are frozen to not affect the duration.

        This is a cached property that will change if the
        corresponding time moment is changed because the item gets
        trimmed or extended.
        """
        if (
            not hasattr(self, '_last_video_frame_cached') or
            self._t_last_video_frame_cached != self._t_last_frame
        ):
            self._t_last_video_frame_cached = self._t_last_frame
            self._last_video_frame_cached = self.get_video_frame_at(
                t_track = self._t_last_video_frame_cached,
                do_use_source_limits = False
            )

        return self._last_video_frame_cached

    # TODO: Why do we have the '__init__' here (?)
    def __init__(
        self,
        is_gap: bool = False,
        # TODO: The position must be the center, based on the center,
        # and (0, 0) is the upper left corner
        video_transform: Union[VideoTransform, None] = None,
        video_effects: Union[VideoEffects, None] = None,
        **kwargs
    ):
        self._is_gap: bool = is_gap
        """
        *For internal use only*

        Internal flag to indicate if this track item is an
        empty part or not.

        TODO: This is a remaining part of the previous way
        to handle the empty parts, so maybe we can remove
        it soon.
        """
        # TODO: What about the time moments (?)
        self.video_transform: Union[VideoTransform, None] = video_transform
        """
        The transformations we want to apply in the video.
        """
        self.video_effects: Union[VideoEffects, None] = video_effects
        """
        The effects we want to apply in the video.
        """

        super().__init__(
            **kwargs
        )

    def get_video_frame_at(
        self,
        t_track: Union[int, float, Fraction],
        do_use_source_limits: bool = False
    ) -> 'VideoFrameWrapped':
        """
        Get the frame that must be displayed at 
        the given 't_track' global time moment.
        """
        frame: Union['VideoFrame', None] = self._get_video_frame_at(t_track, do_use_source_limits)
        # TODO: This can be None, why? I don't know...

        if frame is None:
            ConsolePrinter().print(f'   [ERROR] Frame {str(float(t_track))} was not obtained')

        frame = (
            # I'm using a red full frame to be able to detect
            # fast the frames that were not available, but
            # I need to find the error and find a real solution
            # TODO: This shouldn't happen, its an error
            self._frame_not_found
            if frame is None else
            frame
        )

        # TODO: What about the 'format' (?)
        # TODO: Maybe I shouldn't set the 'time_base'
        # here and do it just in the Timeline 'render'
        #return get_black_background_video_frame(self._track.size)
        # TODO: This 'time_base' maybe has to be related
        # to a Timeline general 'time_base' and not the fps

        # We need to force the track's output size
        """
        TODO: I don't know why but if I comment or uncomment
        this lines, the result is the same. But if I comment
        them in the '_get_video_frame_at' of the 
        'TrackItemWithVideoMedia' it doesn't work.
        
        Check this: https://www.notion.so/Error-duplicidad-VideoTransform-2e1f5a32d462800089ffed20e39a307e?source=copy_link
        """
        # # We need to force the track's output size
        # if self.video_transform is not None:
        #     self.video_transform.output_size = self._track.size

        #     # We need to apply the basic transformations
        #     frame = self.video_transform.apply(
        #         frame = frame,
        #         t = t_track
        #     )

        # TODO: We need to remove the 'is_from_gap' maybe (?)
        frame = VideoFrameWrapped(
            frame = frame,
            is_from_gap = self._is_gap
        )

        # TODO: This should not happen because of
        # the way we handle the videos here but the
        # video could send us a None frame here, so
        # do we raise exception (?)
        if frame._frame is None:
            #frame = get_black_background_video_frame(self._track.size)
            # TODO: By now I'm raising exception to check if
            # this happens or not because I think it would
            # be malfunctioning
            raise Exception(f'Video is returning None video frame at t={str(t_track)}.')
        
        """
        The 'track' (because of the 'timeline' it belongs
        to) has a size, so we must respect, so we need to
        use a strategy to make it fit the desired size.
        """
        # TODO: Maybe this sould be done in the 'timeline'
        # that combines the frames and not here...
        return frame
    
    @abstractmethod
    def _get_video_frame_at(
        self,
        t_track: Union[int, float, Fraction],
        do_use_source_limits: bool = False
    ) -> Union['VideoFrame', None]:
        """
        Get the video frame for the `t_track` time moment
        provided according to the internal configuration of
        this instance.
        """
        pass

class TrackItemWithVideoMedia(_TrackItemWithVideo, _TrackItemWithAudio, _TrackItemWithMedia):
    """
    A track item that is capable of providing video and
    audio frames by obtaining them from the video media
    it contains.
    """

    @property
    def copy(
        self
    ) -> 'TrackItemWithVideoMedia':
        """
        A copy of the current track item with media
        instance.
        """
        return TrackItemWithVideoMedia(
            track = self._track,
            t_start = self.t_start,
            media = self.media,
            audio_transform = (
                self.audio_transform.copy
                if self.audio_transform is not None else
                None
            ),
            video_transform = (
                self.video_transform.copy
                if self.video_transform is not None else
                None
            ),
            audio_effects = (
                self.audio_effects.copy
                if self.audio_effects is not None else
                None
            ),
            video_effects = (
                self.video_effects.copy
                if self.video_effects is not None else
                None
            ),
            item_in = self.item_in,
            item_out = self.item_out,
            transition_in = self.transition_in,
            transition_out = self.transition_out
        )
    
    @property
    def size(
        self
    ) -> tuple[int, int]:
        """
        The size of the media, which is always the original
        size associated to the source of that media.
        """
        return self.media.size
    
    @property
    def output_size(
        self
    ) -> tuple[int, int]:
        """
        The size that we want as the output of this track
        item, which is the output size set into the media,
        that can be a specific one (if provided) or the
        same as the original of the source associated.
        """
        return self.media.output_size
    
    def __init__(
        self,
        track: 'VideoTrack',
        t_start: Union[int, float, Fraction],
        media: 'VideoMedia',
        audio_transform: Union[AudioTransform, None] = None,
        video_transform: Union[VideoTransform, None] = None,
        audio_effects: Union[AudioEffects, None] = None,
        video_effects: Union[VideoEffects, None] = None,
        item_in: Union['_VideoTrackItem', 'TransitionTrackItem', 'GapTrackItem', None] = None,
        item_out: Union['_VideoTrackItem', 'TransitionTrackItem', 'GapTrackItem', None] = None,
        transition_in: Union['TransitionTrackItem', None] = None,
        transition_out: Union['TransitionTrackItem', None] = None,
    ):
        # We force the media size if needed
        if video_transform is not None:
            if video_transform.width == VALUE_UNSET:
                video_transform.width = ConstantVideoEditorParameterValue(media.size[0])
            if video_transform.height == VALUE_UNSET:
                video_transform.height = ConstantVideoEditorParameterValue(media.size[1])
            if video_transform.output_size == VALUE_UNSET:
                video_transform.output_size = track.output_size

        # TODO: Should I send the 'fps' to the media (?)
        super().__init__(
            track = track,
            t_start = t_start,
            media = media,
            audio_transform = audio_transform,
            video_transform = video_transform,
            audio_effects = audio_effects,
            video_effects = video_effects,
            item_in = item_in,
            item_out = item_out,
            transition_in = transition_in,
            transition_out = transition_out,
        )

        # TODO: I think this should be done every time
        # we update the 't_start'
        self._synchronize_video_transform()

    def _synchronize_video_transform(
        self
    ) -> 'TrackItemWithVideoMedia':
        """
        *For internal use only*

        Synchronize the video transform instance by
        setting the `t_start` and `t_end` according to
        this track item time moments.
        """
        # TODO: How do we do this now (?)
        # self.video_transform = (
        #     self.video_transform.update(
        #         t_start = self.t_start,
        #         t_end = self.t_end
        #     )
        #     if self.video_transform is not None else
        #     None
        # )

        return self
    
    # TODO: I think we should overwrite the 'shift_media'
    # and the other methods to adjust the `VideoTransform`
    # if it is needed due to the changes...

    def _get_audio_frames_at(
        self,
        t_track: Union[int, float, Fraction],
        # TODO: Ignore this, we have 'self.fps' here
        video_fps: Union[int, float, Fraction] = None,
        do_use_source_limits: bool = False
    ):
        """
        Get the audio frames for the provided `t_track`
        global time moment from the audio media this instance
        contains.

        The `do_use_source_limits` method should be used
        when we want to access to a `t_track` that is out
        of the time interval of this track item but it is
        actually accepted by the internal source time
        interval limits.
        """
        t_media = self._t_track_to_t_media(
            t_track = t_track,
            do_use_source_limits = do_use_source_limits
        )

        audio_frames = self.media.get_audio_frames_at(
            t_media = t_media,
            video_fps = self.fps,
            do_use_source_limits = do_use_source_limits
        )

        # We need to apply the basic transformations
        audio_frames = (
            [
                # TODO: Is this the way to apply? Individually to the
                # different frames or as a set (?)
                self.audio_transform.apply(
                    frame = audio_frame,
                    t = t_track
                ) for audio_frame in audio_frames
            ]
            if self.audio_transform is not None else
            audio_frames
        )

        # TODO: We need to apply the effects
        audio_frames = (
            [
                # TODO: Is this the way to apply? Individually to the
                # different frames or as a set (?)
                self.audio_effects.apply(
                    frame = audio_frame,
                    t = t_track
                ) for audio_frame in audio_frames
            ]
            if self.audio_effects is not None else
            audio_frames
        )

        return audio_frames

    def _get_video_frame_at(
        self,
        t_track: Union[int, float, Fraction],
        do_use_source_limits: bool = False
    ) -> Union['VideoFrame', None]:
        """
        Get the video frame for the provided `t_track` time
        moment from the video media this instance contains.

        The `do_use_source_limits` method should be used
        when we want to access to a `t_track` that is out
        of the time interval of this track item but it is
        actually accepted by the internal source time
        interval limits.
        """
        t_media = self._t_track_to_t_media(
            t_track = t_track,
            do_use_source_limits = do_use_source_limits
        )

        frame = self.media.get_video_frame_at(
            t_media = t_media,
            do_use_source_limits = do_use_source_limits
        )

        # No modifications? Just return
        if (
            self.video_transform is None and
            self.video_effects is None
        ):
            return frame
        
        t_local = self._t_track_to_t_track_item(t_track)

        # Get the context to be able to apply the changes
        evaluation_context = EvaluationContext.from_t(
            fps = self.fps,
            total_frames = int(self.duration * self.fps),
            t = t_local
        )

        # evaluation_context = self._get_evaluation_context(
        #     t_track = t_track
        # )

        # Transform 'av.VideoFrame' to ndarray to work with it
        frame = frame.to_ndarray(format = 'rgba')

        # We need to force the track's output size
        if self.video_transform is not None:
            # Frame returned will be 'ndarray' (CPU) or 'Texture' (GPU)
            frame = self.video_transform.apply(
                frame = frame,
                # TODO: I'm not sure if the 't' must be the local
                evaluation_context = evaluation_context,
                # TODO: Maybe here we need to provide the duration
                # of the item to be able to calculate
            )

        if self.video_effects is not None:
            # TODO: These effects must be handled different
            # Frame returned will be 'ndarray' (CPU) or 'Texture' (GPU)
            frame = self.video_effects.apply(
                frame = frame,
                output_size = self.output_size,
                evaluation_context = evaluation_context
            )

        # TODO: With this flag we can control it, by now we
        # will be accepting the alpha
        do_include_alpha: bool = True
        format = (
            'rgba'
            if do_include_alpha else
            'rgb24'
        )

        # Frame must be 'av.VideoFrame' when returned
        return VideoFrame.from_ndarray(
            array = frame_to_uint8_ndarray(
                frame = frame,
                do_include_alpha = do_include_alpha
            ),
            format = format
        )

        
"""
Note for the developers:
- We have 3 different types of items:
    *Item part: An item (there are different options) 
    that is occupying that part of the track, being a
    media item, a transition item...
    *Gap part in between: An empty part that is before
    the first media or in between 2 different medias.
    *Gap part at the t_end: An empty part that is at the
    t_end of the track, after the last media, making the
    track virtually infinite.

Check the 'Track' class to see the interaction and
creation.
"""


"""
The audio track module.
"""
from yta_editor.tracks.abstract import _TrackWithAudio


class AudioTrack(_TrackWithAudio):
    """
    Class to represent a track in which we place
    audios to build a video project.
    """

    def __init__(
        self,
        timeline: 'Timeline',
        index: int
    ):
        super().__init__(
            timeline = timeline,
            index = index
        )


from yta_editor.transformations.effects.abstract import Effect


class AudioEffect(Effect):
    """
    An effect for an audio input.

    This class is just to be able to identify the
    hierarchy and which effect is designed for 
    the audio.
    """

    pass

class AudioEffects:
    """
    A set of effects to apply in an audio input.
    """

    # TODO: This is not as simple as a list, it must be
    # managed with the specific nodes we designed that
    # are able to combine and are time-concerned.

    def __init__(
        self,
        effects: list[AudioEffect] = []
    ):
        self.effects: list[AudioEffect] = effects
        """
        The list of the effects to apply in the audio.
        """

    def apply(
        self,
        frame,
        t: float
    # TODO: Set the type
    ) -> any:
        """
        Apply all the effects that should be applied at the
        given `t` time moment to the provided video `frame`.
        """
        for effect in self.effects:
            frame = effect.apply(
                frame = frame,
                t = t
            )

        return frame


from yta_editor.transformations.effects.abstract import EffectParameters
from yta_editor.transformations.effects.video import VideoEffect
from yta_editor_nodes.processor import BrightnessNodeProcessor
from yta_editor_utils.texture import TextureUtils
from yta_editor_time.evaluation_context import EvaluationContext
from yta_editor_parameters.values.abstract import VideoEditorParameterValue
from yta_editor_parameters.values import ConstantVideoEditorParameterValue
from yta_validation.parameter import ParameterValidator
from typing import Union
from dataclasses import dataclass


# TODO: We need to define all the specific effect
# params to be able to build them properly missing
# no fields
@dataclass
class BrightnessVideoEffectParameters(EffectParameters):
    """
    The parameters to use when applying the effect
    that changes the brightness for a specific
    `ealuation_context`.

    This will be returned by the effect when
    calculated, for a specific `evaluation_context`.
    """

    def __init__(
        self,
        brightness: float
    ):
        self.brightness: float = brightness
        """
        The value to apply as brightness.
        """
        

class BrightnessVideoEffect(VideoEffect):
    """
    The effect that will change the brightness of the
    element according to the provided conditions.
    """

    def __init__(
        self,
        do_use_gpu: bool = True,
        # TODO: Use 'factor' instead
        brightness: VideoEditorParameterValue = ConstantVideoEditorParameterValue(2.0)
    ):
        ParameterValidator.validate_mandatory_bool('do_use_gpu', do_use_gpu)
        ParameterValidator.validate_mandatory_instance_of('brightness', brightness, VideoEditorParameterValue)

        self.do_use_gpu: bool = do_use_gpu
        """
        Flag to indicate if using GPU or not.
        """
        # TODO: Maybe 'parse_params' (?)
        self.brightness: VideoEditorParameterValue = brightness
        """
        The `VideoEditorParameterValue` that defines the value that
        should be applied for the specific `evaluation_context`
        requested.
        """

    def _get_params_at(
        self,
        evaluation_context: EvaluationContext
    ) -> BrightnessVideoEffectParameters:
        """
        *For internal use only*

        Get the parameters that must be applied at the given
        `evaluation_context`.
        """
        ParameterValidator.validate_mandatory_instance_of('evaluation_context', evaluation_context, EvaluationContext)

        return BrightnessVideoEffectParameters(
            # TODO: Use 'factor' instead
            brightness = self.brightness.evaluate(
                evaluation_context = evaluation_context
            )
        )
    
    def _apply(
        self,
        # TODO: Set the type
        frame: any,
        output_size: Union[tuple[int, int], None],
        evaluation_context: EvaluationContext
    # TODO: Set the type
    ) -> any:
        parameters = self._get_params_at(
            evaluation_context = evaluation_context
        )

        if parameters.brightness == 1.0:
            return frame
            return TextureUtils.numpy_to_uint8(frame)
        
        return BrightnessNodeProcessor(
            # TODO: What do I do with the 'opengl_context' (?)
            opengl_context = None
        ).process(
            inputs = {
                'base_input': frame
            },
            evaluation_context = evaluation_context,
            do_use_gpu = self.do_use_gpu,
            output_size = output_size,
            # parameters = parameters
            # TODO: Use 'factor' instead
            parameters = {
                'factor': parameters.brightness
            }
        )


"""
TODO: I think this has to be moved from here
"""
from yta_math_easings.enums import EasingFunctionName
from dataclasses import dataclass
from typing import Union


@dataclass
class TransitionSpecification:
    """
    The specification of a transition we will implement,
    including the type, the parameters and everything we
    need to correctly implement it.

    To check the `node_names` that are available, please
    go to the next path:
    - `tracks/items/transition/transition_node_registry.py`
    """

    def __init__(
        self,
        node_name: str,
        parameters: dict,
        easing_function: Union[EasingFunctionName, str] = 'linear'
    ):
        self.node_name: str = node_name
        """
        The name of the node we will use for the transition.
        """
        self.parameters: dict = parameters
        """
        All the parameters we need to correctly instantiate
        the transition node.
        """
        self.easing_function: EasingFunctionName = EasingFunctionName.to_enum(easing_function)
        """
        The easing function we want to apply for the
        progress of the transition, which is `linear`
        by default.
        """

    @staticmethod
    def default(
    ):
        """
        Get the `TransitionSpecification` by default, ready
        to be used in the system.
        """
        from yta_editor.tracks.items.transition.transition_node_registry import TransitionNodeRegistry

        # TODO: Review this, maybe not the best way to have a default
        return TransitionSpecification(
            node_name = TransitionNodeRegistry.get_default_key(),
            parameters = {
                'border_smoothness': 0.02
            },
            easing_function = 'linear'
        )

"""
Considering that 'CircleOpeningTransitionProcessor',
which is a valid and working transition node processor,
has the 'border_smoothness' parameter, we should have
it in the `parameters` field, and the `node_name` should
be `CircleOpeningTransitionProcessor` or similar.

TODO: In a near future we should implement a system
to dynamically validate the parameters we provide to 
the nodes, including type, minimum and maximum value,
etc. Check this code below as inspiration:

--------
@dataclass(frozen=True)
class ParameterSpec:
    name: str
    type: type
    required: bool = True
    default: Any = None
    min: float | None = None
    max: float | None = None

# This is one example of parameters definition
class CircleOpeningTransitionProcessor:
    PARAMS = [
        ParameterSpec("center", tuple),
        ParameterSpec("border_smoothness", float, default=0.0, min=0.0, max=1.0),
        ParameterSpec("invert", bool, default=False),
    ]

# This would be the static validator helper
class TransitionParamValidator:
    @staticmethod
    def validate(specs, params):
        for spec in specs:
            if spec.name not in params:
                if spec.required and spec.default is None:
                    raise ValueError(f"Missing param: {spec.name}")
                params[spec.name] = spec.default

            value = params[spec.name]

            if not isinstance(value, spec.type):
                raise TypeError(f"{spec.name} must be {spec.type}")

            if spec.min is not None and value < spec.min:
                raise ValueError(f"{spec.name} < {spec.min}")

            if spec.max is not None and value > spec.max:
                raise ValueError(f"{spec.name} > {spec.max}")

# We would validate like this:
processor_cls = TransitionProcessorRegistry.get(spec.processor)

TransitionParamValidator.validate(
    processor_cls.PARAMS,
    spec.params
)

# And we could even have animated values that
# depend on the context of the specific frame
# so would be evaluated per frame.
class AnimatedFloat:
    def evaluate(self, ctx) -> float: ...
"""


"""
Manual tests that are working and are interesting
to learn about the code, refactor and build
classes.

TODO: Remove this class when finished and the
call from the '__init__.py' main file.
"""
from yta_logger import ConsolePrinter


IMAGE_FILENAME = 'test_files/mobile_alpha.png'
"""
Filename of an image that has transparency
and its size is 1280x720.
"""

def test_save_image_media():
    from yta_editor.media.video import VideoImageMedia
    from yta_editor.utils import VideoUtils

    output_filename = 'test_files/test_media_image.mov'

    # Transforming an image with alpha layer and
    # transparent pixels into a video with that
    # same transparency
    VideoImageMedia(
        filename = IMAGE_FILENAME,
        duration = 1.0
    ).save_as(
        output_filename = output_filename,
        video_pixel_format = 'argb',
        video_codec = 'qtrle'
    )

    ConsolePrinter().print(VideoUtils.video.alpha.file_has_transparent_pixels(output_filename))

def timeline_to_render():
    from yta_editor.timeline import Timeline
    from yta_editor.media.video import VideoFileMedia, VideoImageMedia

    timeline = Timeline()

    VIDEO_PATH_60FPS = 'test_files/test_1.mp4'
    IMAGE_PATH_A = 'test_files/test_image_1920x1080.webp'
    IMAGE_PATH_B = 'test_files/batman1920.jpg'

    video_60fps_1_5s = VideoFileMedia(
        filename = VIDEO_PATH_60FPS,
        t_start = 0.25,
        t_end = 1.75
    )
    video_60fps_1s = VideoFileMedia(
        filename = VIDEO_PATH_60FPS,
        t_start = 0.25,
        t_end = 1.25
    )
    clip_b = VideoImageMedia(
        filename = IMAGE_PATH_A,
        #duration = 3,
        duration = 1,
        size = video_60fps_1s.size
    )

    video_image_1s_a = VideoImageMedia(
        filename = IMAGE_PATH_A,
        duration = 1,
        size = video_60fps_1s.size
    )
    video_image_1s_b = VideoImageMedia(
        filename = IMAGE_PATH_B,
        duration = 1,
        size = video_60fps_1s.size
    )

    track = timeline.tracks[0]

    track.add_video_item(
        # TODO: We always have to use a 'copy' of the media or
        # the changes we make on them will be applied to the
        # other instances, which is not what we want. We only
        # have to use the same 'source' for the different
        # medias that use it
        media = video_60fps_1s.copy,
        t = 0.33,
        do_force_add = True
    )

    track.add_video_item(
        # TODO: We always have to use a 'copy' of the media or
        # the changes we make on them will be applied to the
        # other instances, which is not what we want. We only
        # have to use the same 'source' for the different
        # medias that use it
        #media = clip_b.copy,
        media = video_60fps_1s.copy,
        t = 1,
        do_force_add = True
    )

    track.add_transition(
        item_in = track.get_item_at(0.4).item,
        item_out = track.get_item_at(1.4).item,
        mode = 'freeze_head',
        duration = 0.5
    )

    timeline.render('test_files/timeline_to_render.mp4')

def test_random_timeline():
    """
    A timeline with different videos created and added
    randomly to be able if the timeline is able to generate
    it properly.
    """
    from yta_editor.timeline import Timeline
    from yta_editor.media.video import VideoFileMedia, VideoImageMedia
    from yta_random import Random
    import random

    timeline = Timeline(number_of_tracks = 1)

    VIDEO_PATH_60FPS = 'test_files/test_1.mp4'
    VIDEO_PATH_30FPS = 'test_files/video_30fps.mp4'
    VIDEO_PATH_29_97FPS = 'test_files/video_29_97fps.mp4'
    IMAGE_PATH_A = 'test_files/test_image_1920x1080.webp'
    IMAGE_PATH_B = 'test_files/batman1920.jpg'

    def random_video(
        filename: str
    ):
        return VideoFileMedia(
            filename = filename,
            t_start = Random.float_between(0.0, 0.75),
            t_end = Random.float_between(1.5, 2.5)
        )
    
    def random_image_video(
        filename: str
    ):
        return VideoImageMedia(
            filename = filename,
            duration = Random.float_between(1.0, 2.0),
            size = (1920, 1080)
        )
    
    for i in range(10):
        track = random.choice(timeline.tracks)
        # track = timeline.tracks[0]
        media = random_video(random.choice([VIDEO_PATH_60FPS, VIDEO_PATH_30FPS, VIDEO_PATH_29_97FPS]))
        # media = random_image_video(random.choice([IMAGE_PATH_A, IMAGE_PATH_B]))
        # media = random.choice([
        #     random_video(random.choice([VIDEO_PATH_60FPS, VIDEO_PATH_30FPS, VIDEO_PATH_29_97FPS])),
        #     random_image_video(random.choice([IMAGE_PATH_A, IMAGE_PATH_B]))
        # ])
        # media = random_video(VIDEO_PATH_60FPS)
        # t = Random.float_between(i*2, i*3)
        t = Random.float_between(i, i)

        ConsolePrinter().print(f'Adding at t={str(float(t))}')
        ConsolePrinter().print(media.__class__.__name__)
        ConsolePrinter().print(media.__str__())

        track.add_video_item(
            media = media,
            t = t,
            do_force_add = True
        )

    # Create random transitions in between the consecutive clips
    from yta_validation import PythonValidator
    for track in timeline.tracks:
        for item in track.items.items:
            # We don't want to add a transition that involves gaps
            if (
                not PythonValidator.is_instance_of(item, ['GapTrackItem', 'TransitionTrackItem']) and
                not PythonValidator.is_instance_of(item.item_out, ['GapTrackItem', 'TransitionTrackItem'])
            ):
                if item.item_out.t_start == item.t_end:
                    track.add_transition(
                        item_in = item,
                        item_out = item.item_out,
                        mode = random.choice(['trim', 'freeze_head', 'freeze_tail', 'freeze_both']),
                        # mode = 'trim',
                        # TODO: Maybe we need to make this duration a multiple of 1/fps
                        #duration = min(item.duration, item.item_out.duration, key = lambda duration: duration) / 2
                        # TODO: What if this number is not even to split in between
                        # the different videos (?)
                        duration = 1/timeline.fps * 24
                    )

    timeline.render('test_files/timeline_to_render_random.mp4')

def video_modified_stored():
    # TODO: Here we can control if printing or
    ConsolePrinter(True)
    from yta_timer import Timer
    with Timer():
        test_random_timeline()
        # timeline_to_render()
    # test_black_frame_cache()

    return
    # TODO: Test a Timeline above this

    # test_opengl_transition()

    # return

    # test_transition()

    # return

    # test_save_image_media()

    # return

    # This path below was trimmed in an online platform
    # and seems to be bad codified and generates error
    # when processing it, but it is readable in the
    # file explorer...
    #VIDEO_PATH = 'test_files/test_1_short_broken.mp4'
    # This is short but is working well
    VIDEO_PATH = 'test_files/test_1_short_2.mp4'
    AUDIO_PATH = 'test_files/test_audio.mp3'
    # Long version below, comment to test faster
    #VIDEO_PATH = 'test_files/test_1.mp4'
    OUTPUT_PATH = 'test_files/output.mp4'
    # TODO: This has to be dynamic, but
    # according to what (?)
    NUMPY_FORMAT = 'rgb24'
    # TODO: Where do we obtain this from (?)
    VIDEO_CODEC_NAME = 'libx264'
    # TODO: Where do we obtain this from (?)
    PIXEL_FORMAT = 'yuv420p'

    from yta_editor.media.video import VideoFileMedia
    from yta_editor.media.audio import AudioFileMedia
    from yta_editor.timeline import Timeline

    # TODO: This test below is just to validate
    # that it is cropping and placing correctly
    # but the videos are only in one track
    # video = Video(VIDEO_PATH, 0.25, 0.75)
    # timeline = Timeline()
    # timeline.add_video(Video(VIDEO_PATH, 0.25, 1.0), 0.5)
    # # This is successfully raising an exception
    # #timeline.add_video(Video(VIDEO_PATH, 0.25, 0.75), 0.6)
    # timeline.add_video(Video(VIDEO_PATH, 0.25, 0.75), 1.75)
    # timeline.add_video(Video('C:/Users/dania/Downloads/Y2meta.app-TOP 12 SIMPLE LIQUID TRANSITION _ GREEN SCREEN TRANSITION PACK-(1080p60).mp4', 4.0, 5.0), 3)
    # # timeline.add_video(Video('C:/Users/dania/Downloads/Y2meta.app-10 Smooth Transitions Green Screen Template For Kinemaster, Alight Motion, Filmora, premiere pro-(1080p).mp4', 2.25, 3.0), 3)
    # timeline.render(OUTPUT_PATH)

    # # Testing concatenating
    # timeline = Timeline()
    # # When you concat like this, some of the
    # # videos have frames that cannot be accessed
    # # and I don't know why...
    # timeline.add_video(Video('test_files/glitch_rgb_frame.mp4'))
    # timeline.add_video(Video('test_files/output.mp4'))
    # timeline.add_video(Video('test_files/output_render.mp4'))
    # timeline.add_video(Video('test_files/strange_tv_frame.mp4'))
    # timeline.add_video(Video('test_files/test_1.mp4'))
    # timeline.add_video(Video('test_files/test_1_short_2.mp4'))
    # timeline.add_video(Video('test_files/test_audio_1st_track_solo_v0_0_15.mp4'))
    # timeline.add_video(Video('test_files/test_audio_2nd_track_solo_v0_0_15.mp4'))
    # timeline.add_video(Video('test_files/test_audio_combined_tracks_v0_0_015.mp4'))
    # timeline.add_video(Video('test_files/test_audio_combined_v0_0_15.mp4'))
    # timeline.add_video(Video('test_files/test_blend_add_v0_0_16.mp4'))
    # timeline.add_video(Video('test_files/test_blend_difference_v0_0_16.mp4'))
    # timeline.add_video(Video('test_files/test_blend_multiply_v0_0_16.mp4'))
    # timeline.add_video(Video('test_files/test_blend_overlay_v0_0_16.mp4'))
    # timeline.add_video(Video('test_files/test_blend_screen_v0_0_16.mp4'))
    # timeline.add_video(Video('test_files/test_combine_skipping_empty_using_priority_v0_0_18.mp4'))
    # timeline.add_video(Video('test_files/test_ok_v0_0_13.mp4'))

    # timeline.render('test_files/concatenated.mp4')

    # from yta_video_pyav.media import ImageMedia, ColorMedia

    # image_media = ImageMedia('test_files/mobile_alpha.png', 0, 1).save_as('test_files/test_image.mp4')

    # color_media = ColorMedia('random', 0, 1).save_as('test_files/test_color.mp4')

    # return

    # TODO: This test will add videos that
    # must be played at the same time

    # from yta_video_ffmpeg.handler import FfmpegHandler

    # ffmpeg_handler = FfmpegHandler()

    # # ffmpeg_handler.video.encoding.to_dnxhr(VIDEO_PATH, 'test_files/video_dnxhr.mov')
    # # ffmpeg_handler.video.encoding.to_prores(VIDEO_PATH, 'test_files/video_prores.mov')
    # # ffmpeg_handler.video.encoding.to_mjpeg(VIDEO_PATH, 'test_files/video_mjpeg.mov')

    # ffmpeg_handler.video.trim_fast(VIDEO_PATH, 0.25, 0.75, 'test_files/trimmed_fast.mp4')
    # ffmpeg_handler.video.trim_accurate(VIDEO_PATH, 0.25, 0.75, 'test_files/trimmed_accurate.mp4')
    # ffmpeg_handler.audio.trim_fast(AUDIO_PATH, 0.25, 1.75, 'test_files/trimmed_fast.mp3')
    # ffmpeg_handler.audio.trim_accurate(AUDIO_PATH, 0.25, 1.75, 'test_files/trimmed_accurate.mp3')

    # return


    # audio = Audio(AUDIO_PATH, 3, 6).save_as('test_files/output.mp3')

    # return

    # TODO: Testing overlaying with alpha layer

    # import av
    # for name in av.codecs_available:
    #     ConsolePrinter().print(name)
    # exit()
    # import av
    # qtrle = av.codec.Codec('qtrle', 'r')  # 'r' = decode/read
    # ConsolePrinter().print("is_decoder:", qtrle.is_decoder)
    # ConsolePrinter().print("is_encoder:", qtrle.is_encoder)
    # ConsolePrinter().print("type:", qtrle.type)
    # exit()

    # TODO: Transform non-alpha transition into
    # a alpha transition
    def non_alpha_transition_to_alpha_transition():
        from yta_video_pyav.writer import VideoWriter
        from yta_video_frame_time.t_fraction import get_ts

        video_alpha = VideoFileMedia('test_files/alpha_transition.mp4')

        writer = VideoWriter('test_files/alpha_transition.mov')
        # 'prores_ks' and 'yuva444p10le' write but not seen

        """
        To transform a non-alpha transition in 
        an alpha transition, use this:
        - codec: 'libvpx-vp9'
        - extension: '.webm'
        - pixel_format: 'yuva420p'
        """

        """
        'libvpx-vp9' and 'yuva420p' with '.webm'
        is written but even when the array has an
        alpha layer, the result is 'yuv420p' with
        the alpha metadata, but not including it.
        """

        """
        'qtrle' and 'argb' with '.mov' is working
        and writting the video correctly with the
        alpha layer, but then I cannot read it
        properly...
        """
        
        writer.set_video_stream(
            codec_name = 'qtrle', # video_alpha.codec_name, # 'libx264'
            fps = video_alpha.fps,
            size = video_alpha.size,
            pixel_format = 'argb'
        )

        # TODO: By now I'm omitting the audio
        # writer.set_audio_stream(
        #     codec_name = self.audio_codec,
        #     fps = self.audio_fps
        # )
        import numpy as np
        from av.video.frame import VideoFrame

        for t in get_ts(0, video_alpha.t_end, video_alpha.fps):
            frame = video_alpha.get_video_frame_at_t(t)

            # TODO: Why isn't this frame available if
            # I can see it in the original one (?)
            if frame is None:
                ConsolePrinter().print(f'Frame at "t={str(t)}" is None (not available).')
                continue

            # To RGB
            rgb = frame.to_ndarray(format = 'rgb24')

            threshold = 30

            # Maximum brightness per pixel
            # brightness = np.max(rgb, axis=-1)  # valor más alto de R,G,B
            # mask = (brightness > threshold).astype(np.uint8) * 255
            # alpha = mask[..., None]

            # Pure black only
            # Create alpha: 0 when black, 255 when other
            # mask = np.any(rgb != [0, 0, 0], axis = -1).astype(np.uint8) * 255
            # alpha = mask[..., None]  # (H, W, 1)

            # With euclidean distance
            dist = np.linalg.norm(rgb.astype(np.int16), axis=-1)
            mask = (dist > threshold).astype(np.uint8) * 255
            alpha = mask[..., None]

            # To RGBA
            rgba = np.concatenate([rgb, alpha], axis = -1)

            # Create to write
            out_frame = VideoFrame.from_ndarray(rgba, format = 'rgba')
            #out_frame = out_frame.reformat(format = 'yuva420p')
            #out_frame.pict_type = 'NONE'

            # Codificar y muxear
            writer.mux_video_frame(
                frame = out_frame
            )

        writer.mux_video_frame(None)
        writer.output.close()

    #non_alpha_transition_to_alpha_transition()

    #return

    """
    Reading frames from .mov directly
    """
    # import av
    # import numpy as np

    # # Abrimos el archivo .mov con qtrle
    # container = av.open("test_files/alpha_transition.mov")

    # # Seleccionamos el primer stream de video
    # video_stream = container.streams.video[0]

    # for frame in container.decode(video_stream):
    #     # Forzamos conversión del frame a 'rgba'
    #     rgba_frame = frame.reformat(format="rgba")

    #     # Lo pasamos a un numpy array (H, W, 4)
    #     img = rgba_frame.to_ndarray()

    #     ConsolePrinter().print(img.shape, img.dtype)  # Ej: (480, 640, 4) uint8

    # exit()

    # IMAGE_WITH_ALPHA = 'test_files/mobile_alpha.png'

    # from yta_editor.media.video import VideoImageMedia, VideoColorMedia

    # timeline = Timeline(
    #     video_codec = 'qtrle',
    #     video_pixel_format = 'argb'
    # )

    # #video_alpha = VideoFileMedia('test_files/alpha_transition_euclidean_distance_30.mov')
    # video_alpha = VideoImageMedia(IMAGE_WITH_ALPHA, 1, size = (1280, 720), do_include_alpha = True)
    # # I manually force the video_alpha size
    # # that has that size due to the image
    # #video_image = VideoImageMedia('test_files/background_1920_1080.jpg', 1, size = (1280, 720), do_include_alpha = True)
    # video_color = VideoColorMedia('notworkingyet', 1, size = (1280, 720), transparency = 0.1)

    # timeline.add_video(video_alpha, 0.2)
    # timeline.add_video(video_color, 0.0, 1)

    # timeline.render(OUTPUT_PATH.replace('.mp4', '.mov'), output_format = 'argb')

    # return

    def transition_to_alpha(
        filename: str,
        output_filename: str
    ):
        """
        Save the provided transition 'filename' file
        to a video file with the alpha layer.

        Perfect to download from AlphaPediaYT and
        transform into clips with alpha, using the
        chromakey, and using them as transitions.
        """
        from yta_video_pyav.reader.filter.dataclass import GraphFilters

        video = VideoFileMedia(
            filename = filename,
            video_filters = [
                GraphFilters.video.chromakey(
                    hex_color = '0x000000'
                )
            ]
        )

        video.save_as(
            output_filename = output_filename,
            video_size = video.size,
            video_fps = video.fps,
            #video_codec = video.codec_name, # that supports alpha
            video_codec = 'prores_ks',
            #video_pixel_format = video.pixel_format, # with alpha
            video_pixel_format = 'yuva444p10le',
            audio_codec = video.audio_codec_name,
            audio_sample_rate = video.audio_fps,
            audio_format = video.audio_format,
            audio_layout = video.audio_layout
        )

    # transition_to_alpha(
    #     filename = 'test_files/alpha_crocodile_transition.mp4',
    #     output_filename = 'test_files/alpha_transition_automated.mov'
    # )

    # return

    test_timeline_simpsons()

    return

    # Track 1
    timeline.add_video(Video(VIDEO_PATH, 0.25, 1.0), 0.75, track_index = 0)
    timeline.add_video(Video(simpsons_60fps, 1.5, 2.0), 3.0, track_index = 0)
    timeline.add_video(Video(VIDEO_PATH, 0.5, 1.0), 2.0, track_index = 0)

    #timeline.tracks[0].mute()

    # Track 2
    timeline.add_video(Video(VIDEO_PATH, 0.5, 1.0), 2.7, track_index = 1)
    timeline.add_video(Video(simpsons_60fps, 5.8, 7.8), 0.6, track_index = 1)
    # 30fps
    # timeline.add_video(Video('C:/Users/dania/Downloads/Y2meta.app-TOP 12 SIMPLE LIQUID TRANSITION _ GREEN SCREEN TRANSITION PACK-(1080p60).mp4', 0.25, 1.5), 0.25, do_use_second_track = True)
    # 29.97fps
    # timeline.add_video(Video('C:/Users/dania/Downloads/Y_una_porra_los_simpsons_castellano.mp4', 5.8, 6.8), 3.6, do_use_second_track = True)
    
    timeline.render(OUTPUT_PATH)

    return

    Video(VIDEO_PATH, 0.25, 0.75).save_as(OUTPUT_PATH)

    return


from yta_editor.utils import VideoUtils
from yta_validation.parameter import ParameterValidator
from av.video.frame import VideoFrame
from av.audio.frame import AudioFrame
from typing import Union

import numpy as np


IS_FROM_GAP_METADATA = 'is_from_gap'
"""
Metadata key to indicate if the frame
has been generated by a gap and should
be ignored when trying to combine with
others.
"""
IS_SILENT_METADATA = 'is_silent'
"""
Metadata key to indicate if the frame is a
silent frame or not, being able to handle it
properly when we need to combine it.
"""

class _FrameWrappedBase:
    """
    Class to wrap video and audio frames from
    the pyav library but to support a metadata
    field to inject some information we need
    when processing and combining them.
    """

    @property
    def frame(
        self
    ) -> VideoFrame:
        """
        Get the original pyav VideoFrame.
        """
        return self._frame

    @property
    def is_from_gap(
        self
    ) -> bool:
        """
        Flag to indicate if the frame comes from
        an empty part or not, that will be done
        by checking the 'is_from_gap' 
        attribute in the metadata.

        This will return True only if the value
        is set and it is set as 'True'.
        """
        return self._is_metadata_boolean_set(IS_FROM_GAP_METADATA)
    
    @property
    def is_silent(
        self
    ) -> bool:
        """
        Flag to indicate if the frame that comes is
        actually an empty frame or not, which could
        be useful when we need to combine it with 
        others, and let us avoid reading the content
        to detect if it is silent or not.

        This metadata is valid for audio frames.

        This will return `True` only if the value is
        set and its value is `'True'`.
        """
        return self._is_metadata_boolean_set(IS_SILENT_METADATA)

    def __init__(
        self,
        frame: Union[VideoFrame, AudioFrame],
        metadata: dict = {}
    ):
        ParameterValidator.validate_mandatory_instance_of('frame', frame, [VideoFrame, AudioFrame])
        ParameterValidator.validate_mandatory_dict('metadata', metadata)

        self._frame: Union[VideoFrame, AudioFrame] = frame
        """
        The VideoFrame or AudioFrame pyav instance.
        """
        self.metadata: dict = metadata or {}
        """
        The metadata we want to include with the
        frame.
        """

    def __getattr__(
        self,
        name
    ):
        return getattr(self._frame, name)

    def __setattr__(
        self,
        name,
        value
    ):
        super().__setattr__(name, value)
        #setattr(self._frame, name, value)
        # if name in ('_frame', 'metadata'):
        #     super().__setattr__(name, value)
        # else:
        #     setattr(self._frame, name, value)

    def __repr__(
        self
    ):
        cname = self.__class__.__name__

        return f'<{cname} metadata={self.metadata} frame={self._frame!r}>'
    
    def set_is_from_gap(
        self,
        value: bool = True
    ) -> None:
        """
        Add the metadata information to indicate
        that this is a frame that comes from an 
        empty part.
        """
        return self._set_metadata_boolean(
            attribute_name = IS_FROM_GAP_METADATA,
            value = value
        )

    def set_is_silent(
        self,
        value: bool = True
    ) -> None:
        """
        Add the metadata information to indicate that this
        is a silent frame, including the `is_silent` 
        metadata.
        """
        return self._set_metadata_boolean(
            attribute_name = IS_SILENT_METADATA,
            value = value
        )

    def _is_metadata_boolean_set(
        self,
        attribute_name: str
    ) -> bool:
        """
        *For internal use only*

        Method to simplify the way we check if the boolean
        metadata `attribute_name` is set or not. A metadata
        boolean parameter must be stored as string `'True`'
        to be considered valid.
        """
        is_parameter = self.metadata.get(attribute_name, None)

        return (
            is_parameter == 'True'
            if is_parameter is not None else
            is_parameter
        )
    
    def _set_metadata_boolean(
        self,
        attribute_name: str,
        value: bool = True
    ) -> None:
        """
        Set the `attribute_name` metadata boolean parameter
        `value` provided (that will be stored as a str).
        """
        self.metadata[attribute_name] = (
            'True'
            if value else
            'False'
        )
    
class VideoFrameWrapped(_FrameWrappedBase):
    """
    Class to wrap video frames from the pyav
    library but to support a metadata field
    to inject some information we need when
    processing and combining them.
    """

    @property
    def has_alpha_layer(
        self
    ) -> bool:
        """
        Flag to indicate if the frame has an alpha
        layer or not, that is defined by an 'a' in
        its format name.
        """
        return VideoUtils.video.alpha.videoframe_has_alpha_layer(self._frame)
    
    @property
    def has_alpha_pixels(
        self
    ) -> bool:
        """
        Flag to indicate if the frame comes with
        at least one pixel that has some kind of
        transparency and has to be considered for
        the alpha blending process because of that.
        """
        return (
            self.has_alpha_layer and
            VideoUtils.video.alpha.numpy_videoframe_has_transparent_pixels(self.as_rgba_numpy)
        )
    
    @property
    def size(
        self
    ) -> tuple[int, int]:
        """
        The size of the frame expressed as (width,
        height).
        """
        return (self.frame.width, self.frame.height)

    @property
    def as_rgb_numpy(
        self
    ) -> 'np.ndarray':
        """
        Get the video frame but as a numpy array with
        the 'rgb24' format, that doesn't include the
        alpha channel.

        The 'rgb24' format is perfect for video
        edition including not the alpha channel.
        """
        return self.frame.to_ndarray(format = 'rgb24')
    
    @property
    def as_rgba_numpy(
        self
    ) -> 'np.ndarray':
        """
        Get the video frame but as a numpy array with
        the 'rgba' format, that includes the alpha
        channel.

        The 'rgba' format is perfect for video edition
        including the alpha channel.
        """
        return self.frame.to_ndarray(format = 'rgba')

    def __init__(
        self,
        frame: VideoFrame,
        metadata: dict = {},
        is_from_gap: Union[bool, None] = None,
    ):
        ParameterValidator.validate_mandatory_instance_of('frame', frame, VideoFrame)

        super().__init__(
            frame = frame,
            metadata = metadata
        )

        if is_from_gap is not None:
            self.set_is_from_gap(is_from_gap)
        
# TODO: Maybe move? But careful with
# cyclic import issues
class AudioFrameWrapped(_FrameWrappedBase):
    """
    Class to wrap audio frames from the pyav
    library but to support a metadata field
    to inject some information we need when
    processing and combining them.
    """

    @property
    def copy(
        self
    ) -> 'AudioFrameWrapped':
        """
        A copy of this same instance.
        """
        return AudioFrameWrapped(
            # TODO: Do I need to hard-copy this '_frame' (?)
            frame = self._frame,
            metadata = self.metadata
        )

    def __init__(
        self,
        frame: AudioFrame,
        metadata: dict = {},
        is_from_gap: bool = False,
        is_silent: bool = False
    ):
        ParameterValidator.validate_mandatory_instance_of('frame', frame, AudioFrame)

        super().__init__(
            frame = frame,
            metadata = metadata
        )

        if is_from_gap:
            self.set_is_from_gap()

        if is_silent:
            self.set_is_silent()


"""
The video track module.
"""
from yta_editor.tracks.abstract import _TrackWithAudio, _TrackWithVideo


class VideoTrack(_TrackWithVideo, _TrackWithAudio):
    """
    Class to represent a track in which we place
    videos to build a video project.
    """

    def __init__(
        self,
        timeline: 'Timeline',
        index: int
    ):
        super().__init__(
            timeline = timeline,
            index = index
        )